#!/usr/bin/env python3
"""
Aided regularization parameter estimation.

@author: Nicola VIGANÒ, Computational Imaging group, CWI, The Netherlands,
and ESRF - The European Synchrotron, Grenoble, France
"""

import concurrent.futures as cf
import inspect
import multiprocessing as mp
import time as tm
from abc import ABC, abstractmethod
from typing import Any, Callable, Optional, Sequence, Union, Literal
from warnings import warn

import matplotlib.pyplot as plt
import numpy as np
from numpy.polynomial.polynomial import Polynomial
from numpy.typing import ArrayLike, DTypeLike, NDArray
from matplotlib.ticker import StrMethodFormatter
from tqdm.auto import tqdm

from . import solvers

num_threads = round(np.log2(mp.cpu_count() + 1))


NDArrayFloat = NDArray[np.floating]


def create_random_test_mask(
    data_shape: Sequence[int], test_fraction: float = 0.05, dtype: DTypeLike = np.float32
) -> NDArrayFloat:
    """
    Create a random mask for cross-validation.

    Parameters
    ----------
    data_shape : Sequence[int]
        The shape of the data.
    test_fraction : float, optional
        The fraction of pixels to mask. The default is 0.05.
    data_dtype : DTypeLike, optional
        The data type of the mask. The default is np.float32.

    Returns
    -------
    data_test_mask : NDArrayFloat
        The pixel mask.
    """
    data_test_mask = np.zeros(data_shape, dtype=dtype)
    num_test_pixels = int(np.ceil(data_test_mask.size * test_fraction))
    test_pixels = np.random.permutation(data_test_mask.size)
    test_pixels = np.unravel_index(test_pixels[:num_test_pixels], data_shape)
    data_test_mask[test_pixels] = 1
    return data_test_mask


def get_lambda_range(start: float, end: float, num_per_order: int = 4, aligned_order: bool = True) -> NDArrayFloat:
    """Compute hyper-parameter values within an interval.

    Parameters
    ----------
    start : float
        First hyper-parameter value.
    end : float
        Last hyper-parameter value.
    num_per_order : int, optional
        Number of steps per order of magnitude. The default is 4.
    aligned_order : bool, optional
        Whether to align the 1 of each order of magnitude or to the given start value. The default is True.

    Returns
    -------
    NDArrayFloat
        List of hyper-parameter values.
    """
    step_size = 10 ** (1 / num_per_order)
    if aligned_order:
        order_start = 10 ** np.floor(np.log10(start))
        order_end = 10 ** np.ceil(np.log10(end))
        num_steps = np.ceil(num_per_order * np.log10(order_end / order_start) - 1e-3)
        tmp_steps = order_start * (step_size ** np.arange(num_steps + 1))
        return tmp_steps[np.logical_and(tmp_steps >= start, (tmp_steps * (1 - 1e-3)) <= end)]
    else:
        num_steps = np.ceil(num_per_order * np.log10(end / start) - 1e-3)
        return start * (step_size ** np.arange(num_steps + 1))


def fit_func_min(
    hp_vals: Union[ArrayLike, NDArrayFloat],
    f_vals: NDArrayFloat,
    f_stds: Optional[NDArrayFloat] = None,
    scale: Literal["linear", "log"] = "log",
    verbose: bool = False,
    plot_result: bool = False,
) -> tuple[float, float]:
    """Parabolic fit of objective function costs for the different hyper-parameter values.

    Parameters
    ----------
    hp_vals : Union[ArrayLike, NDArrayFloat]
        Hyper-parameter values.
    f_vals : NDArrayFloat
        Objective function costs of each hyper-parameter value.
    f_stds : NDArrayFloat, optional
        Objective function cost standard deviations of each hyper-parameter value.
        It is only used for plotting purposes. The default is None.
    scale : str, optional
        Scale of the fit. Options are: "log" | "linear". The default is "log".
    verbose : bool, optional
        Whether to produce verbose output, by default False
    plot_result : bool, optional
        Whether to plot the result, by default False

    Returns
    -------
    min_hp_val : float
        Expected hyper-parameter value of the fitted minimum.
    min_f_val : float
        Expected objective function cost of the fitted minimum.
    """
    hp_vals = np.array(hp_vals, ndmin=1)

    if len(hp_vals) < 3 or len(f_vals) < 3 or len(hp_vals) != len(f_vals):
        raise ValueError(
            "Lengths of the lambdas and function values should be identical and >= 3."
            f"Given: lams={len(hp_vals)}, vals={len(f_vals)}"
        )

    if scale.lower() == "log":
        to_fit = np.log10
        from_fit = lambda x: 10**x
    elif scale.lower() == "linear":
        to_fit = lambda x: x
        from_fit = to_fit
    else:
        raise ValueError(f"Parameter 'scale' should be either 'log' or 'linear', given '{scale}' instead")

    min_pos = np.argmin(f_vals)
    if min_pos == 0:
        warn("Minimum value at the beginning of the lambda range.")
        hp_inds_fit = list(np.arange(3))
    elif min_pos == (len(f_vals) - 1):
        warn("Minimum value at the end of the lambda range.")
        hp_inds_fit = list(np.mod(np.arange(-3, 0), hp_vals.size))
    else:
        hp_inds_fit = list(np.arange(min_pos - 1, min_pos + 2))
    lams_reg_fit = to_fit(hp_vals[hp_inds_fit])
    f_vals_fit = f_vals[hp_inds_fit]

    counter = tm.perf_counter()
    if verbose:
        print(
            f"Fitting minimum within the interval [{hp_vals[hp_inds_fit[0]]:.3e}, {hp_vals[hp_inds_fit[-1]]:.3e}]"
            f" (indices: [{hp_inds_fit[0]}, {hp_inds_fit[-1]}]): ",
            end="",
            flush=True,
        )

    # using Polynomial.fit, because it is supposed to be more numerically
    # stable than previous solutions (according to numpy).
    poly = Polynomial.fit(lams_reg_fit, f_vals_fit, deg=2)
    coeffs = poly.convert().coef
    if coeffs[2] <= 0:
        warn("Fitted curve is concave. Returning minimum measured point.")
        return hp_vals[min_pos], f_vals[min_pos]

    # For a 1D parabola `f(x) = c + bx + ax^2`, the vertex position is:
    # x_v = -b / 2a.
    vertex_pos = -coeffs[1] / (2 * coeffs[2])
    vertex_val = coeffs[0] + vertex_pos * coeffs[1] / 2

    min_hp_val, min_f_val = from_fit(vertex_pos), vertex_val
    if min_hp_val < hp_vals[0] or min_hp_val > hp_vals[-1]:
        warn(
            f"Fitted lambda {min_hp_val:.3e} is outside the bounds of input lambdas [{hp_vals[0]:.3e}, {hp_vals[-1]:.3e}]."
            " Returning minimum measured point."
        )
        res_hp_val, res_f_val = hp_vals[min_pos], f_vals[min_pos]
    else:
        res_hp_val, res_f_val = min_hp_val, min_f_val

    if verbose:
        print(f"Found at {min_hp_val:.3e}, in {tm.perf_counter() - counter:g} seconds.\n")

    if plot_result:
        fig, axs = plt.subplots()
        axs.set_xscale(scale, nonpositive="clip")
        if f_stds is None:
            axs.plot(hp_vals, f_vals)
        else:
            axs.errorbar(hp_vals, f_vals, yerr=f_stds, ecolor=(0.5, 0.5, 0.5), elinewidth=1, capsize=2)
        x = np.linspace(lams_reg_fit[0], lams_reg_fit[2])
        y = coeffs[0] + x * (coeffs[1] + x * coeffs[2])
        axs.plot(from_fit(x), y)
        axs.scatter(min_hp_val, min_f_val)
        axs.grid()
        for tl in axs.get_xticklabels():
            tl.set_fontsize(13)
        for tl in axs.get_yticklabels():
            tl.set_fontsize(13)
        axs.set_xlabel(r"$\lambda$ values", fontsize=16)
        axs.set_ylabel("Cross-validation loss values", fontsize=16)
        axs.yaxis.set_major_formatter(StrMethodFormatter("{x:.2e}"))
        fig.tight_layout()
        plt.show(block=False)

    return res_hp_val, res_f_val


class BaseParameterTuning(ABC):
    """Base class for parameter tuning classes."""

    _solver_spawning_functionls: Optional[Callable]
    _solver_calling_function: Optional[Callable[[Any], tuple[NDArrayFloat, solvers.SolutionInfo]]]

    def __init__(
        self, dtype: DTypeLike = np.float32, parallel_eval: bool = True, verbose: bool = False, plot_result: bool = False
    ) -> None:
        """Initialize a base helper class.

        Parameters
        ----------
        dtype : DTypeLike, optional
            Type of the data, by default np.float32
        parallel_eval : bool, optional
            Whether to evaluate results in parallel, by default True
        verbose : bool, optional
            Whether to produce verbose output, by default False
        plot_result : bool, optional
            Whether to plot the results, by default False
        """
        self.dtype = dtype

        self.parallel_eval = parallel_eval
        self.verbose = verbose
        self.plot_result = plot_result

        self._solver_spawning_function = None
        self._solver_calling_function = None

    @property
    def solver_spawning_function(self) -> Callable:
        """Return the locally stored solver spawning function."""
        if self._solver_spawning_function is None:
            raise ValueError("Solver spawning function not initialized!")
        return self._solver_spawning_function

    @property
    def solver_calling_function(self) -> Callable[[Any, ...], tuple[NDArrayFloat, solvers.SolutionInfo]]:
        """Return the locally stored solver calling function."""
        if self._solver_calling_function is None:
            raise ValueError("Solver spawning function not initialized!")
        return self._solver_calling_function

    @solver_spawning_function.setter
    def solver_spawning_function(self, solver_spawn: Callable) -> None:
        if not isinstance(solver_spawn, Callable):
            raise ValueError("Expected a solver spawning function (callable)")
        if len(inspect.signature(solver_spawn).parameters) != 1:
            raise ValueError(
                "Expected a solver spawning function (callable), whose only parameter is the regularization lambda"
            )
        self._solver_spawning_function = solver_spawn

    @solver_calling_function.setter
    def solver_calling_function(self, solver_call: Callable[[Any, ...], tuple[NDArrayFloat, solvers.SolutionInfo]]) -> None:
        if not isinstance(solver_call, Callable):
            raise ValueError("Expected a solver calling function (callable)")
        if not len(inspect.signature(solver_call).parameters) >= 1:
            raise ValueError("Expected a solver calling function (callable), with at least one parameter (solver)")
        self._solver_calling_function = solver_call

    @staticmethod
    def get_lambda_range(start: float, end: float, num_per_order: int = 4) -> NDArrayFloat:
        """Compute regularization weights within an interval.

        Parameters
        ----------
        start : float
            First regularization weight.
        end : float
            Last regularization weight.
        num_per_order : int, optional
            Number of steps per order of magnitude. The default is 4.

        Returns
        -------
        NDArrayFloat
            List of regularization weights.
        """
        return get_lambda_range(start=start, end=end, num_per_order=num_per_order)

    def compute_reconstruction_and_loss(self, hp_val: float, *args: Any, **kwds: Any) -> tuple[float, NDArrayFloat]:
        """Compute objective function cost for the given hyper-parameter value.

        Parameters
        ----------
        hp_val : float
            hyper-parameter value.
        *args : Any
            Optional positional arguments for the reconstruction.
        **kwds : Any
            Optional keyword arguments for the reconstruction.

        Returns
        -------
        cost : float
            Cost of the given regularization weight.
        rec : NDArray
            Reconstruction at the given weight.
        """
        solver = self.solver_spawning_function(hp_val)
        rec, rec_info = self.solver_calling_function(solver, *args, **kwds)

        # Output will be: test objective function cost, and reconstruction
        return float(rec_info.residuals_cv_rel[-1]), rec

    def compute_reconstruction_error(
        self, hp_vals: Union[ArrayLike, NDArrayFloat], gnd_truth: NDArrayFloat
    ) -> tuple[NDArrayFloat, NDArrayFloat]:
        """Compute the reconstruction errors for each hyper-parameter values against the ground truth.

        Parameters
        ----------
        hp_vals : Union[ArrayLike, NDArrayFloat]
            List of hyper-parameter values.
        gnd_truth : NDArrayFloat
            Expected reconstruction.

        Returns
        -------
        err_l1 : NDArrayFloat
            l1-norm errors for each reconstruction.
        err_l2 : NDArrayFloat
            l2-norm errors for each reconstruction.
        """
        hp_vals = np.array(hp_vals, ndmin=1)

        if self.verbose:
            print("Computing reconstruction error:")
            print(f"- Hyper-parameter values range: [{hp_vals[0]:.3e}, {hp_vals[-1]:.3e}] in {len(hp_vals)} steps")
            print(f"Parallel evaluation: {self.parallel_eval} (n. threads: {num_threads})")

        if self.parallel_eval:
            with cf.ThreadPoolExecutor(max_workers=num_threads) as executor:
                future_to_lambda = {
                    executor.submit(self.compute_reconstruction_and_loss, l): (ii, l) for ii, l in enumerate(hp_vals)
                }

                recs = [np.array([])] * len(hp_vals)
                try:
                    for future in tqdm(
                        cf.as_completed(future_to_lambda),
                        desc="Hyper-parameter values",
                        disable=not self.verbose,
                        total=len(hp_vals),
                    ):
                        hp_ind, hp_val = future_to_lambda[future]
                        try:
                            recs[hp_ind] = future.result()[1]
                        except ValueError as exc:
                            print(f"Hyper-parameter value {hp_val} (#{hp_ind}) generated an exception: {exc}")
                            raise
                except:
                    print("Shutting down..", end="", flush=True)
                    executor.shutdown(cancel_futures=True)
                    print("\b\b: Done.")
                    raise
        else:
            recs = [
                self.compute_reconstruction_and_loss(l)[1]
                for l in tqdm(hp_vals, desc="Hyper-parameter values", disable=not self.verbose)
            ]

        err_l1 = np.zeros((len(hp_vals),), dtype=self.dtype)
        err_l2 = np.zeros((len(hp_vals),), dtype=self.dtype)

        for ii_l, rec in enumerate(recs):
            residual = np.abs(gnd_truth - rec)
            err_l1[ii_l] = np.linalg.norm(residual.ravel(), ord=1)
            err_l2[ii_l] = np.linalg.norm(residual.ravel(), ord=2) ** 2

        if self.plot_result:
            fig, axs = plt.subplots(2, 1, sharex=True)
            axs[0].set_xscale("log", nonpositive="clip")  # type: ignore
            axs[0].plot(hp_vals, err_l1, label="Error - l1-norm")  # type: ignore
            axs[0].legend()  # type: ignore
            axs[1].set_xscale("log", nonpositive="clip")  # type: ignore
            axs[1].plot(hp_vals, err_l2, label="Error - l2-norm ^ 2")  # type: ignore
            axs[1].legend()  # type: ignore
            fig.tight_layout()
            plt.show(block=False)

        return err_l1, err_l2

    @abstractmethod
    def compute_loss_values(self, hp_vals: Union[ArrayLike, NDArrayFloat]) -> NDArrayFloat:
        """Compute the objective function costs for a list of hyper-parameter values.

        Parameters
        ----------
        hp_vals : Union[ArrayLike, NDArrayFloat]
            List of hyper-parameter values.

        Returns
        -------
        NDArrayFloat
            Objective function cost for each hyper-parameter value.
        """


class LCurve(BaseParameterTuning):
    """L-curve regularization parameter estimation helper."""

    def __init__(
        self,
        loss_function: Callable,
        data_dtype: DTypeLike = np.float32,
        parallel_eval: bool = True,
        verbose: bool = False,
        plot_result: bool = False,
    ):
        """Create an L-curve regularization parameter estimation helper.

        Parameters
        ----------
        loss_function : Callable
            The loss function for the computation of the L-curve values.
        data_dtype : DTypeLike, optional
            Type of the input data. The default is np.float32.
        parallel_eval : bool, optional
            Compute loss and error values in parallel. The default is False.
        verbose : bool, optional
            Print verbose output. The default is False.
        plot_result : bool, optional
            Plot results. The default is False.

        Raises
        ------
        ValueError
            In case 'loss_function' is not callable or does not expose at least one argument.
        """
        super().__init__(dtype=data_dtype, parallel_eval=parallel_eval, verbose=verbose, plot_result=plot_result)

        if not isinstance(loss_function, Callable):
            raise ValueError(
                "Expected a callable with one argument for the argument 'loss_function',"
                " whose parameters are: the solver and the data test mask"
            )
        if len(inspect.signature(loss_function).parameters) != 1:
            raise ValueError("The callable 'loss_function', should have one parameter")
        self.loss_function = loss_function

    def compute_loss_values(self, hp_vals: Union[ArrayLike, NDArrayFloat]) -> NDArrayFloat:
        """Compute objective function values for all hyper-parameter values.

        Parameters
        ----------
        hp_vals : Union[ArrayLike, NDArrayFloat]
            Hyper-parameter values to use for computing the different objective function values.

        Returns
        -------
        f_vals : NDArrayFloat
            Objective function cost for each hyper-parameter value.
        """
        hp_vals = np.array(hp_vals, ndmin=1)

        counter = tm.perf_counter()
        if self.verbose:
            print("Computing L-curve loss values:")
            print(f"- Hyper-parameter values range: [{hp_vals[0]:.3e}, {hp_vals[-1]:.3e}] in {len(hp_vals)} steps")
            print(f"Parallel evaluation: {self.parallel_eval} (n. threads: {num_threads})")

        if self.parallel_eval:
            with cf.ThreadPoolExecutor(max_workers=num_threads) as executor:
                future_to_lambda = {
                    executor.submit(self.compute_reconstruction_and_loss, l): (ii, l) for ii, l in enumerate(hp_vals)
                }

                recs = [np.array([])] * len(hp_vals)
                try:
                    for future in tqdm(
                        cf.as_completed(future_to_lambda),
                        desc="Hyper-parameter values",
                        disable=not self.verbose,
                        total=len(hp_vals),
                    ):
                        hp_ind, hp_val = future_to_lambda[future]
                        try:
                            recs[hp_ind] = future.result()[1]
                        except ValueError as exc:
                            print(f"Hyper-parameter value {hp_val} (#{hp_ind}) generated an exception: {exc}")
                            raise
                except:
                    print("Shutting down..", end="", flush=True)
                    executor.shutdown(cancel_futures=True)
                    print("\b\b: Done.")
                    raise
        else:
            recs = [
                self.compute_reconstruction_and_loss(l)[1]
                for l in tqdm(hp_vals, desc="Hyper-parameter values", disable=not self.verbose)
            ]

        f_vals = np.array([self.loss_function(rec) for rec in recs], dtype=self.dtype)

        if self.verbose:
            print(f"Done in {tm.perf_counter() - counter} seconds.\n")

        if self.plot_result:
            fig, axs = plt.subplots()
            axs.set_title("L-Curve loss values")
            axs.set_xscale("log", nonpositive="clip")
            axs.set_yscale("log", nonpositive="clip")
            axs.plot(hp_vals, f_vals)
            axs.grid()
            fig.tight_layout()
            plt.show(block=False)

        return f_vals


class CrossValidation(BaseParameterTuning):
    """Cross-validation hyper-parameter estimation class."""

    def __init__(
        self,
        data_shape: Sequence[int],
        dtype: DTypeLike = np.float32,
        cv_fraction: float = 0.1,
        num_averages: int = 7,
        parallel_eval: bool = True,
        verbose: bool = False,
        plot_result: bool = False,
    ):
        """Create a cross-validation hyper-parameter estimation helper.

        Parameters
        ----------
        data_shape : Sequence[int]
            Shape of the projection data.
        data_dtype : DTypeLike, optional
            Type of the input data. The default is np.float32.
        cv_fraction : float, optional
            Fraction of detector points to use for the leave-out set. The default is 0.1.
        num_averages : int, optional
            Number of averages random leave-out sets to use. The default is 7.
        parallel_eval : bool, optional
            Compute loss and error values in parallel. The default is False.
        verbose : bool, optional
            Print verbose output. The default is False.
        plot_result : bool, optional
            Plot results. The default is False.
        """
        super().__init__(dtype=dtype, parallel_eval=parallel_eval, verbose=verbose, plot_result=plot_result)
        self.data_shape = data_shape
        self.cv_fraction = cv_fraction
        self.num_averages = num_averages

        self.data_test_masks = [self._create_random_test_mask() for _ in range(self.num_averages)]

    def _create_random_test_mask(self) -> NDArrayFloat:
        return create_random_test_mask(self.data_shape, self.cv_fraction, self.dtype)

    def compute_loss_values(self, hp_vals: Union[ArrayLike, NDArrayFloat]) -> tuple[NDArrayFloat, NDArrayFloat, NDArrayFloat]:
        """Compute objective function values for all requested hyper-parameter values..

        Parameters
        ----------
        params : Union[ArrayLike, NDArrayFloat]
            Hyper-parameter values (e.g. regularization weight) to evaluate.

        Returns
        -------
        f_avgs : NDArrayFloat
            Average objective function costs for each regularization weight.
        f_stds : NDArrayFloat
            Standard deviation of objective function costs for each regularization weight.
        f_vals : NDArrayFloat
            Objective function costs for each regularization weight.
        """
        hp_vals = np.array(hp_vals, ndmin=1)

        counter = tm.perf_counter()
        if self.verbose:
            print("Computing cross-validation loss values:")
            print(f"- Hyper-parameter range: [{hp_vals[0]:.3e}, {hp_vals[-1]:.3e}] in {len(hp_vals)} steps")
            print(f"- Number of averages: {self.num_averages}")
            print(f"- Leave-out pixel fraction: {self.cv_fraction:%}")
            print(f"Parallel evaluation: {self.parallel_eval} (n. threads: {num_threads})")

        f_vals = np.empty((len(hp_vals), self.num_averages), dtype=self.dtype)
        for ii_avg in range(self.num_averages):
            if self.verbose:
                print(f"\nRound: {ii_avg + 1}/{self.num_averages}")

            if self.parallel_eval:
                with cf.ThreadPoolExecutor(max_workers=num_threads) as executor:
                    future_to_lambda = {
                        executor.submit(self.compute_reconstruction_and_loss, l, self.data_test_masks[ii_avg]): (ii, l)
                        for ii, l in enumerate(hp_vals)
                    }

                    f_vals_ii = [0.0] * len(hp_vals)
                    try:
                        for future in tqdm(
                            cf.as_completed(future_to_lambda),
                            desc="Hyper-parameter values",
                            disable=not self.verbose,
                            total=len(hp_vals),
                        ):
                            hp_ind, hp_val = future_to_lambda[future]
                            try:
                                f_vals_ii[hp_ind] = future.result()[0]
                            except ValueError as exc:
                                print(f"Hyper-parameter value {hp_val} (#{hp_ind}) generated an exception: {exc}")
                                raise
                    except:
                        print("Shutting down..", end="", flush=True)
                        executor.shutdown(cancel_futures=True)
                        print("\b\b: Done.")
                        raise
            else:
                f_vals_ii = [
                    self.compute_reconstruction_and_loss(l, self.data_test_masks[ii_avg])[0]
                    for l in tqdm(hp_vals, desc="Hyper-parameter values", disable=not self.verbose)
                ]

            f_vals[:, ii_avg] = np.array(f_vals_ii)

        f_avgs = np.mean(f_vals, axis=1)
        f_stds = np.std(f_vals, axis=1)

        if self.verbose:
            print(f"Done in {tm.perf_counter() - counter:g} seconds.\n")

        if self.plot_result:
            fig, axs = plt.subplots()
            axs.set_title(f"Cross-validation loss values (avgs: {self.num_averages})")
            axs.set_xscale("log", nonpositive="clip")
            axs.errorbar(hp_vals, f_avgs, yerr=f_stds, ecolor=(0.5, 0.5, 0.5), elinewidth=1, capsize=2)
            for f in f_vals.T:
                axs.plot(hp_vals, f, linewidth=1, linestyle="--")
            axs.grid()
            fig.tight_layout()
            plt.show(block=False)

        return f_avgs, f_stds, f_vals

    def fit_loss_min(
        self,
        hp_vals: Union[ArrayLike, NDArrayFloat],
        f_vals: NDArrayFloat,
        f_stds: Optional[NDArrayFloat] = None,
        scale: Literal["linear", "log"] = "log",
    ) -> tuple[float, float]:
        """Parabolic fit of objective function costs for the different hyper-parameter values.

        Parameters
        ----------
        hp_vals : Union[ArrayLike, NDArrayFloat]
            Hyper-parameter values.
        f_vals : NDArrayFloat
            Objective function costs of each hyper-parameter value.
        f_stds : NDArrayFloat, optional
            Objective function cost standard deviations of each hyper-parameter value.
            It is only used for plotting purposes. The default is None.
        scale : str, optional
            Scale of the fit. Options are: "log" | "linear". The default is "log".

        Returns
        -------
        min_hp_val : float
            Expected hyper-parameter value of the fitted minimum.
        min_f_val : float
            Expected objective function cost of the fitted minimum.
        """
        return fit_func_min(
            hp_vals=hp_vals, f_vals=f_vals, f_stds=f_stds, scale=scale, verbose=self.verbose, plot_result=self.plot_result
        )



"""
Advanced denoising methods.

@author: Nicola VIGANÒ, Computational Imaging group, CWI, The Netherlands,
and ESRF - The European Synchrotron, Grenoble, France
"""

from collections.abc import Sequence
from typing import Callable, Optional, Union, overload

import numpy as np
import scipy.signal as spsig
from numpy.typing import NDArray

from . import data_terms, operators, param_tuning, regularizers, solvers

eps = np.finfo(np.float32).eps


def _default_regularizer_l1dwl(r_w: Union[float, NDArray]) -> regularizers.BaseRegularizer:
    return regularizers.Regularizer_l1dwl(r_w, "bior4.4", 3)


@overload
def denoise_image(
    img: NDArray,
    reg_weight: Union[Sequence[float], NDArray],
    psf: Optional[NDArray] = None,
    pix_weights: Optional[NDArray] = None,
    iterations: int = 250,
    regularizer: Callable = _default_regularizer_l1dwl,
    lower_limit: Optional[float] = None,
    verbose: bool = True,
) -> tuple[NDArray, float]: ...


@overload
def denoise_image(
    img: NDArray,
    reg_weight: float,
    psf: Optional[NDArray] = None,
    pix_weights: Optional[NDArray] = None,
    iterations: int = 250,
    regularizer: Callable = _default_regularizer_l1dwl,
    lower_limit: Optional[float] = None,
    verbose: bool = True,
) -> NDArray: ...


def denoise_image(
    img: NDArray,
    reg_weight: Union[float, Sequence[float], NDArray] = 1e-2,
    psf: Optional[NDArray] = None,
    pix_weights: Optional[NDArray] = None,
    iterations: int = 250,
    regularizer: Callable = _default_regularizer_l1dwl,
    lower_limit: Optional[float] = None,
    verbose: bool = True,
) -> Union[NDArray, tuple[NDArray, float]]:
    """
    Denoise an image.

    Image denoiser based on (flat or weighted) least-squares, with wavelet minimization regularization.
    The weighted least-squares requires the local pixel-wise weights.
    It can be used to denoise sinograms and projections.

    Parameters
    ----------
    img : NDArray
        The image to denoise.
    reg_weight : Union[float, ArrayLike, NDArray], optional
        Weight of the regularization term. The default is 1e-2.
        If a sequence / array is passed, all the different values will be tested.
        The one minimizing the error over the cross-validation set will be chosen and returned.
    pix_weights : Union[ArrayLike, NDArray, None], optional
        The local weights of the pixels, for a weighted least-squares minimization.
        If None, a standard least-squares minimization is performed. The default is None.
    iterations : int, optional
        Number of iterations. The default is 250.
    regularizer : Callable, optional
        The one-argument constructor of a regularizer. The default is the DWL regularizer.
    lower_limit : Optional[float], optional
        Lower clipping limit of the image. The default is None.
    verbose : bool, optional
        Turn verbosity on. The default is True.

    Returns
    -------
    NDArray
        Denoised image.
    """
    if psf is None:
        op = operators.TransformIdentity(img.shape)
    else:
        op = operators.TransformConvolution(img.shape, psf)

    if pix_weights is None:
        data_term = data_terms.DataFidelity_l2()
    else:
        data_term = data_terms.DataFidelity_wl2(pix_weights)

    def solver_spawn(lam_reg):
        # Using the PDHG solver from Chambolle and Pock
        reg = regularizer(lam_reg)
        return solvers.PDHG(
            verbose=verbose, data_term=data_term, regularizer=reg, data_term_test=data_term, leave_progress=False
        )

    def solver_call(solver: solvers.Solver, b_test_mask: Optional[NDArray] = None) -> tuple[NDArray, solvers.SolutionInfo]:
        x0 = img.copy()
        if b_test_mask is not None:
            med_img = spsig.medfilt2d(img, kernel_size=11)
            masked_pixels = b_test_mask > 0.5

            x0[masked_pixels] = med_img[masked_pixels]

        return solver(op, img, iterations, x0=x0, lower_limit=lower_limit, b_test_mask=b_test_mask)

    reg_weight = np.array(reg_weight)
    if reg_weight.size > 1:
        reg_help_cv = param_tuning.CrossValidation(img.shape, verbose=verbose, num_averages=3, plot_result=verbose)
        reg_help_cv.solver_spawning_function = solver_spawn
        reg_help_cv.solver_calling_function = solver_call

        f_avgs, _, _ = reg_help_cv.compute_loss_values(reg_weight)

        min_reg_weight, _ = reg_help_cv.fit_loss_min(reg_weight, f_avgs)
    else:
        min_reg_weight = reg_weight

    solver = solver_spawn(min_reg_weight)
    denoised_img, _ = solver_call(solver, None)

    if reg_weight.size == 1:
        return denoised_img
    else:
        return denoised_img, min_reg_weight



#!/usr/bin/env python3
"""
Provide structured illumination support.

Created on Sun Jan  9 17:39:02 2022

@author: Nicola VIGANÒ, Computational Imaging group, CWI, The Netherlands,
and ESRF - The European Synchrotron, Grenoble, France
"""

import copy as cp

from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import Union

import matplotlib.pyplot as plt
import numpy as np
import scipy.linalg as spalg

from numpy.typing import DTypeLike, NDArray

from tqdm.auto import tqdm

from . import operators, processing

try:
    import torch as pt

    __has_torch_cuda__ = pt.cuda.is_available()
except ImportError:
    __has_torch_cuda__ = False


NDArrayInt = NDArray[np.integer]


def reorder_masks(masks: NDArray, buckets: NDArray, shift: int) -> tuple[NDArray, NDArray]:
    """Reorder masks, with a simple rot-n algorithm.

    Parameters
    ----------
    masks : NDArray
        The masks to re-order.
    buckets : NDArray
        The corresponding buckets.
    shift : int
        The length of the shift.

    Returns
    -------
    Tuple[NDArray, NDArray]
        The reordered masks and buckets.
    """
    masks_shape = masks.shape
    masks = masks.reshape([-1, *masks_shape[-2:]])

    masks = np.roll(masks, -shift, axis=0)
    buckets = np.roll(buckets, -shift)

    return masks.reshape(masks_shape), buckets


def decompose_qr_masks(masks: NDArray, verbose: bool = False) -> tuple[NDArray, NDArray]:
    """Compute QR decomposition of the given masks.

    Parameters
    ----------
    masks : NDArray
        The masks to decompose
    verbose : bool, optional
        Whether to emite verbose output, by default False

    Returns
    -------
    Tuple[NDArray, NDArray]
        The Q and R components.
    """
    masks_shape = masks.shape
    masks = masks.reshape([-1, np.prod(masks_shape[-2:])])

    if verbose:
        print("Computing QR decomposition")
    Q, R = np.linalg.qr(masks.transpose())

    Rt = R.transpose()
    Qt = Q.transpose()
    if verbose:
        print("Computing inversion of R (fast, because triangular)")
    R1t = np.linalg.inv(Rt)

    return Qt.reshape(masks_shape), R1t


def estimate_resolution(masks: NDArray, verbose: bool = True, plot_result: bool = True) -> tuple[float, float]:
    """Estimate the mask collection resolution through auto-correlation.

    Parameters
    ----------
    masks : NDArray
        The list of encoding masks
    verbose : bool, optional
        Whether to produce verbose output, by default True
    plot_result : bool, optional
        Whether to plot the results, by default True

    Returns
    -------
    tuple[float, float]
        The mean and minimum HWHM of the auto-correlation functions for all the masks.
    """
    masks = masks.reshape([-1, *masks.shape[-2:]])

    masks = masks - masks.mean(axis=(-2, -1), keepdims=True)

    resolutions = np.zeros(len(masks))
    for ind, mask in enumerate(tqdm(masks, desc="Computing auto-correlations", disable=not verbose)):
        _, auto_corr = processing.misc.norm_cross_corr(mask, plot=False)
        point = processing.misc.lines_intersection(auto_corr, 0.5, position="first")
        resolutions[ind] = point[0] if point is not None else 0

    res_mean = resolutions.mean()
    res_min = resolutions.min()

    if plot_result:
        fig, axs = plt.subplots(1, 1, figsize=[9, 3.75])
        axs.plot(resolutions, label="HWHM auto-correlation")
        axs.hlines(res_mean, 0, len(resolutions), colors=["C1"], label=f"Mean: {res_mean:.3} pix")
        axs.hlines(res_min, 0, len(resolutions), colors=["C2"], label=f"Min: {res_min:.3} pix")
        axs.grid()
        axs.legend(fontsize=13)
        axs.tick_params(labelsize=16)
        fig.tight_layout()
        plt.show(block=False)

    return res_mean, res_min


class MaskCollection:
    """Define mask collection class."""

    masks_enc: NDArray
    masks_dec: NDArray

    mask_dims: int
    mask_support: NDArrayInt

    mask_type: str

    def __init__(
        self,
        masks_enc: NDArray,
        masks_dec: Union[NDArray, None] = None,
        mask_dims: int = 2,
        mask_type: str = "measured",
        mask_support: Union[None, Sequence[int], NDArrayInt] = None,
    ) -> None:
        """Initialize mask collection.

        Parameters
        ----------
        masks_enc : NDArray
            The encoding masks.
        masks_dec : NDArray | None, optional
            The decoding masks. The default (None) will assume them identical to the encoding masks.
        mask_dims : int, optional
            The dimensions of a single mask. The default is 2.
        mask_type : str, optional
            The type of masks. The default is "measured".
        mask_support : Sequence[int] | NDArray[np.integer] | None, optional
            The extent of the mask support in pixels. The default (None) will consider it equal to the FoV.
        """
        self.mask_dims = mask_dims

        if masks_enc.ndim < mask_dims + 1:
            raise ValueError(f"Encoding masks should have at least {mask_dims + 1} dimensions")

        if masks_dec is not None:
            if masks_dec.ndim < mask_dims + 1:
                raise ValueError(f"Decoding masks should have at least {mask_dims + 1} dimensions")

            if masks_enc.shape[:-mask_dims] != masks_dec.shape[:-mask_dims]:
                raise ValueError("Encoding and decoding masks should have the same shape for the stack dimensions")

            if masks_enc.shape[-mask_dims:] != masks_dec.shape[-mask_dims:]:
                raise ValueError("Encoding and decoding masks should have the same shape for each mask")

        self.masks_enc = masks_enc
        self.masks_dec = masks_dec if masks_dec is not None else masks_enc

        self.mask_type = mask_type

        # TODO: check support is honored
        if mask_support is not None:
            self.mask_support = np.array(mask_support, ndmin=1, dtype=int)
        else:
            self.mask_support = np.array(self.shape_fov, ndmin=1, dtype=int)

    @property
    def shape_fov(self) -> Sequence[int]:
        """Return the mask shape.

        Returns
        -------
        Sequence[int]
            The mask shape.
        """
        return self.masks_enc.shape[-self.mask_dims :]

    @property
    def shape_shifts(self) -> Sequence[int]:
        """Compute the shape of the available shifts.

        Returns
        -------
        Sequence[int]
            The shape of the available shifts.
        """
        return self.masks_enc.shape[: -self.mask_dims]

    @property
    def num_buckets(self) -> int:
        """Compute the total number of available buckets.

        Returns
        -------
        int
            The total number of buckets.
        """
        return int(np.prod(self.shape_shifts))

    @property
    def num_pixels(self) -> int:
        """Compute the number of pixels in the image.

        Returns
        -------
        int
            The number of pixels in the image.
        """
        return int(np.prod(self.shape_fov))

    def info(self) -> str:
        """
        Return the mask info.

        Returns
        -------
        str
            Mask info string.
        """
        return self.mask_type

    def upper(self) -> str:
        """
        Return the upper case name of the mask.

        Returns
        -------
        str
            Upper case string name of the mask.
        """
        return self.mask_type.upper()

    def lower(self) -> str:
        """
        Return the lower case name of the mask.

        Returns
        -------
        str
            Lower case string name of the mask.
        """
        return self.mask_type.lower()

    def get_mask(self, mask_inds_vu: Union[Sequence, NDArray], mask_encoding: bool = True) -> NDArray:
        """Return the requested mask.

        Parameters
        ----------
        mask_inds_vu : Union[Sequence, NDArray]
            The mask position.
        mask_encoding : bool, optional
            Whether it is an encoding or decoding mask, by default True

        Returns
        -------
        NDArray
            The requested mask.
        """
        if mask_encoding is True:
            mask = self.masks_enc
        else:
            mask = self.masks_dec

        return mask[tuple(mask_inds_vu)]

    def get_QR_decomposition(self, buckets: NDArray, shift: int = 0) -> tuple["MaskCollection", NDArray]:
        """Compute and return the QR decomposition of the masks.

        Parameters
        ----------
        buckets : NDArray
            The buckets corresponding to the masks.
        shift : int, optional
            Index of the first mask, by default 0.

        Returns
        -------
        Tuple[MaskCollection, NDArray]
            The new mask collection and the matrix for modifying buckets accordingly.

        Raises
        ------
        ValueError
            In case the masks have encoding-decoding form.
        """
        if np.any(self.masks_dec != self.masks_enc):
            raise ValueError("Cannot compute QR decomposition of masks that have encoding and decoding forms.")

        new_masks = cp.deepcopy(self)
        new_buckets = cp.deepcopy(buckets)
        if shift != 0:
            new_masks.masks_enc, new_buckets = reorder_masks(new_masks.masks_enc, new_buckets, shift=shift)

        new_masks.masks_enc, R1t = decompose_qr_masks(new_masks.masks_enc)
        new_masks.masks_dec = new_masks.masks_enc
        new_buckets = R1t.dot(new_buckets)

        return new_masks, new_buckets

    def bin_masks(self, binning: float) -> "MaskCollection":
        """Bin the masks.

        Parameters
        ----------
        binning : float
            The binning size.

        Returns
        -------
        MaskCollection
            A new collection of binned masks.
        """
        new_masks = cp.deepcopy(self)

        new_masks.masks_enc = processing.pre.bin_imgs(new_masks.masks_enc, binning=binning)
        new_masks.masks_dec = processing.pre.bin_imgs(new_masks.masks_dec, binning=binning)

        return new_masks

    def inspect_masks(self, mask_inds_vu: Union[None, Sequence[int], NDArrayInt] = None):
        """Inspect the encoding and decoding masks at the requested shifts.

        Parameters
        ----------
        mask_inds_vu : Sequence[int] | NDArray[np.integer] | None, optional
            The requested axes shifts. The default (None) is the first mask.
        """
        if mask_inds_vu is None:
            mask_inds_vu = [0] * len(self.shape_shifts)

        mask_enc = self.get_mask(mask_inds_vu, mask_encoding=True)
        mask_dec = self.get_mask(mask_inds_vu, mask_encoding=False)

        f, axs = plt.subplots(1, 2)
        f.suptitle("Base masks")

        axs[0].imshow(mask_enc)
        axs[0].set_title("Encoding")
        axs[1].imshow(mask_dec)
        axs[1].set_title("Decoding")

        plt.show(block=False)


class MaskGenerator(ABC):
    """Define mask generation interface."""

    shape_fov: NDArrayInt
    shape_mask: NDArrayInt
    shape_shifts: NDArrayInt
    transmittance: float
    dtype: DTypeLike

    _enc_dec_mismatch: bool

    __mask_name__ = "generated"

    def __init__(
        self,
        shape_fov: Union[Sequence[int], NDArrayInt],
        shape_mask: Union[Sequence[int], NDArrayInt],
        shape_shifts: Union[Sequence[int], NDArrayInt],
        transmittance: float = 1.0,
        dtype: DTypeLike = np.float32,
    ) -> None:
        """Initialize mask collection.

        Parameters
        ----------
        shape_fov : Sequence[int] | NDArray[np.integer]
            The shape of the field-of-view.
        shape_mask : Sequence[int] | NDArray[np.integer]
            The shape of the masks.
        shape_shifts : Sequence[int] | NDArray[np.integer]
            The shape of the shifts to generate.
        transmittance : float
            The maximum transmittance of the structuring elements.
        dtype : DTypeLike
            The dtype of the created masks.
        """
        self.shape_fov = np.array(shape_fov, dtype=int)
        self.shape_mask = np.array(shape_mask, dtype=int)
        self.shape_shifts = np.array(shape_shifts, dtype=int)

        self.transmittance = transmittance

        self.dtype = dtype

        self._enc_dec_mismatch = False

    def info(self) -> str:
        """
        Return the mask info.

        Returns
        -------
        str
            Mask info string.
        """
        return self.__mask_name__

    def __repr__(self) -> str:
        """Produce the string representation of the object.

        Returns
        -------
        str
            The string representation.
        """
        return self.__class__.__name__ + " {\n" + ",\n".join([f"  {k} = {v}" for k, v in self.__dict__.items()]) + "\n}"

    @property
    def max_buckets(self) -> int:
        """Compute the maximum number of buckets.

        Returns
        -------
        int
            The maximum number of buckets.
        """
        return int(np.prod(self.shape_shifts))

    @property
    def num_pixels(self) -> int:
        """Compute the number of pixels in the image.

        Returns
        -------
        int
            The number of pixels in the image.
        """
        return int(np.prod(self.shape_fov))

    def _init_fov_mm(self, fov_size_mm: Union[float, Sequence[float], NDArray], req_res_mm: float) -> NDArrayInt:
        self.fov_size_mm = np.array(fov_size_mm, ndmin=1)
        num_points = np.ceil(self.fov_size_mm / req_res_mm).astype(int)
        self.feature_size_mm = self.fov_size_mm / num_points

        return num_points

    def generate_collection(self, buckets_fraction: float = 1, shift_type: str = "sequential") -> MaskCollection:
        """Generate the mask collection.

        Parameters
        ----------
        buckets_fraction : float, optional
            The fraction of buckets to generate, by default 1
        shift_type : str, optional
            The type of shift to implement, by default "sequential"
        abs_fraction : float, optional
            The attenuation fraction of the pixels

        Returns
        -------
        MaskCollection
            The generated mask collection.

        Raises
        ------
        ValueError
            In case of wrong shift type.
        """
        if shift_type.lower() == "random":
            num_chosen_buckets = np.ceil(self.num_pixels * buckets_fraction).astype(int)
            disp_v, disp_u = self.get_random_shifts(num_chosen_buckets)
        elif shift_type.lower() == "interval":
            interval = np.ceil(1 / buckets_fraction).astype(int)
            disp_v, disp_u = self.get_interval_shifts(interval)
            num_chosen_buckets = len(disp_v)
        elif shift_type.lower() == "sequential":
            num_chosen_buckets = np.ceil(self.num_pixels * buckets_fraction).astype(int)
            disp_v, disp_u = self.get_sequential_shifts(num_chosen_buckets)
        else:
            raise ValueError(f'Wrong shift_type: "{shift_type}". Available options: random | interval | sequential.')

        gen_masks_enc = self._generate_mask_shifts(disp_v, disp_u, mask_encoding=True)
        gen_masks_dec = self._generate_mask_shifts(disp_v, disp_u, mask_encoding=False) if self._enc_dec_mismatch else None

        print(
            f"Using {num_chosen_buckets} masks/buckets over {self.max_buckets} (max available), and {self.num_pixels} pixels"
        )

        return MaskCollection(
            gen_masks_enc,
            gen_masks_dec,
            mask_type=self.__mask_name__,
            mask_support=self.shape_mask,
            mask_dims=len(self.shape_fov),
        )

    def _apply_transmission(self, masks: NDArray) -> NDArray:
        return 1 - (1 - masks) * self.transmittance

    @abstractmethod
    def generate_shifted_mask(self, mask_inds_vu: Union[Sequence, NDArray], mask_encoding: bool = True) -> NDArray:
        """Produce the shifted masks.

        Parameters
        ----------
        mask_inds_vu : tuple | list | NDArray
            The vertical and horizontal shifts.
        mask_encoding : bool, optional
            Is the mask encoding (False = decoding). The default is True.

        Returns
        -------
        NDArray
            The shifted mask.
        """

    def _generate_mask_shifts(
        self, shifts_v: Union[Sequence, NDArray], shifts_u: Union[Sequence, NDArray], mask_encoding: bool = True
    ) -> NDArray:
        """Produce all the masks.

        Parameters
        ----------
        shifts_v : tuple | list | NDArray
            List of all the vertical shifts.
        shifts_u : tuple | list | NDArray
            List of all the horizontal shifts.
        mask_encoding : bool, optional
            Are the masks encoding (False = decoding). The default is True.
        abs_fraction : float, optional
            Absorption fraction of the mask elements. The default is 1.

        Returns
        -------
        NDArray
            The collection of all the shifted masks.
        """
        masks = [self.generate_shifted_mask([v, u], mask_encoding) for v, u in zip(shifts_v, shifts_u)]
        return np.stack(masks, axis=0)  # .reshape([*self.shift_shape, *self.fov_shape])

    def get_interval_shifts(
        self, interval: Union[int, Sequence[int], NDArray], axes_order: Sequence[int] = (-2, -1)
    ) -> Sequence[NDArray]:
        """Produce shifts for the "interval" shift type.

        Parameters
        ----------
        interval : int | tuple(int, int) | list(int, int)
            The shift interval.
        axes_order : int | tuple | list, optional
            Order of the axes to shift. The default is (-2, -1).

        Returns
        -------
        tuple
            The collection of shifts.
        """
        interval = np.array(interval, dtype=int)
        if interval.size == 1:
            interval = np.tile(interval, len(axes_order))

        disps = [np.arange(0, self.shape_shifts[ax], interval[ii]) for ii, ax in enumerate(axes_order)]
        disps = np.meshgrid(*disps, indexing="ij")
        return [disp.flatten() for disp in disps]

    def get_random_shifts(self, num_shifts: int, axes_order: Sequence[int] = (-2, -1)) -> Sequence[NDArray]:
        """Produce shifts for the "random" shift type.

        Parameters
        ----------
        num_shifts : int
            Number of shifts.

        Returns
        -------
        NDArray
            The collection of shifts.
        """
        if num_shifts > self.max_buckets:
            print(f"Warning, too many shifts. Truncating to: {self.max_buckets}")
        num_shifts = np.fmin(num_shifts, self.max_buckets)

        disps = self.get_interval_shifts(interval=1, axes_order=axes_order)
        perms = np.random.permutation(np.prod(self.shape_shifts))[:num_shifts]
        return [disp[perms] for disp in disps]

    def get_sequential_shifts(
        self, num_shifts: Union[int, None] = None, axes_order: Sequence[int] = (-2, -1)
    ) -> Sequence[NDArray]:
        """Produce shifts for the "sequential" shift type.

        Parameters
        ----------
        num_shifts : int, optional
            Number of shifts. The default is None.
        axes_order : tuple | list | NDArray, optional
            Order of the axes to shift. The default is (-2, -1).

        Returns
        -------
        NDArray
            The collection of shifts.
        """
        disps = self.get_interval_shifts(interval=1, axes_order=axes_order)
        if num_shifts is not None:
            disps = [disp[:num_shifts] for disp in disps]
        return disps


class MaskGeneratorPoint(MaskGenerator):
    """Pencil beam masks generator class."""

    __mask_name__ = "pencil"

    def __init__(self, fov_size_mm: Union[float, Sequence[float], NDArray], req_res_mm: float = 1.0):
        """Initialize the pencil beam mask collection.

        Parameters
        ----------
        fov_size_mm : float
            Size of the Field-of-View in millimiters.
        req_res_mm : float
            Requested resolution in millimiters.
        """
        num_points = self._init_fov_mm(fov_size_mm, req_res_mm)
        super().__init__(num_points, [1, 1], num_points)

    def generate_shifted_mask(self, mask_inds_vu: Union[Sequence, NDArray], mask_encoding: bool = True) -> NDArray:
        """Produce the shifted masks.

        Parameters
        ----------
        mask_inds_vu : tuple | list | NDArray
            The vertical and horizontal shifts.
        mask_encoding : bool, optional
            Is the mask encoding (False = decoding). The default is True.

        Returns
        -------
        NDArray
            The shifted mask.
        """
        mask = np.zeros(self.shape_fov, dtype=self.dtype)
        mask[mask_inds_vu[0], mask_inds_vu[1]] = 1.0
        return self._apply_transmission(mask)


class MaskGeneratorBernoulli(MaskGenerator):
    """Bernoulli mask generator class."""

    __mask_name__ = "bernoulli"

    def __init__(
        self, fov_size_mm: Union[float, Sequence[float], NDArray], req_res_mm: float = 1.0, max_masks_ratio: float = 1.2
    ):
        """
        Bernulli masks collection class.

        It computes and stores the original mask pattern for a given resolution and Field-ofView.

        Parameters
        ----------
        fov_size_mm : float
            Size of the Field-of-View in mm.
        req_res_mm : float
            The required pixel size in mm.
        max_masks_ratio : float
            The ratio of maximum available masks per pixels in each direction, by default 1.2.
        """
        num_points = self._init_fov_mm(fov_size_mm, req_res_mm)
        super().__init__(
            shape_fov=num_points, shape_mask=num_points, shape_shifts=np.ceil(num_points * max_masks_ratio).astype(int)
        )

    def generate_shifted_mask(self, mask_inds_vu: Union[Sequence, NDArray], mask_encoding: bool = True) -> NDArray:
        """Produce the shifted masks.

        Parameters
        ----------
        mask_inds_vu : tuple | list | NDArray
            The vertical and horizontal shifts.
        mask_encoding : bool, optional
            Is the mask encoding (False = decoding). The default is True.

        Returns
        -------
        NDArray
            The shifted mask.
        """
        mask = np.random.randint(0, 2, size=self.shape_fov).astype(self.dtype)
        return self._apply_transmission(mask)


class MaskGeneratorHalfGaussian(MaskGenerator):
    """Half-Gaussian mask generator class."""

    __mask_name__ = "half-gaussian"

    def __init__(
        self, fov_size_mm: Union[float, Sequence[float], NDArray], req_res_mm: float = 1.0, max_masks_ratio: float = 1.2
    ):
        """
        Half Gaussian masks collection class.

        It computes and stores the original mask pattern for a given resolution and Field-ofView.

        Parameters
        ----------
        fov_size_mm : float
            Size of the Field-of-View in mm.
        req_res_mm : float
            The required pixel size in mm.
        max_masks_ratio : float
            The ratio of maximum available masks per pixels in each direction, by default 1.2.
        """
        num_points = self._init_fov_mm(fov_size_mm, req_res_mm)
        super().__init__(
            shape_fov=num_points, shape_mask=num_points, shape_shifts=np.ceil(num_points * max_masks_ratio).astype(int)
        )

    def generate_shifted_mask(self, mask_inds_vu: Union[Sequence, NDArray], mask_encoding: bool = True) -> NDArray:
        """Produce the shifted masks.

        Parameters
        ----------
        mask_inds_vu : tuple | list | NDArray
            The vertical and horizontal shifts.
        mask_encoding : bool, optional
            Is the mask encoding (False = decoding). The default is True.

        Returns
        -------
        NDArray
            The shifted mask.
        """
        mask = np.abs(np.random.randn(*self.shape_fov)).astype(self.dtype)
        return self._apply_transmission(mask)


class MaskGeneratorMURA(MaskGenerator):
    """MURA mask generator class."""

    __mask_name__ = "mura"

    def __init__(self, fov_size_mm: float, req_res_mm: float = 1.0):
        """
        MURA masks collection class.

        Parameters
        ----------
        fov_size_mm : float
            Size of the Field-of-View in mm.
        req_res_mm : float
            The required pixel size in mm.
        """
        self.fov_size_mm = np.array([fov_size_mm, fov_size_mm])
        base_points = int(np.ceil((fov_size_mm / req_res_mm - 1) / 4))
        num_points = 4 * base_points + 1
        self.feature_size_mm = self.fov_size_mm / num_points

        sq = np.mod(np.arange(num_points) ** 2, num_points)

        mask_1d = np.zeros(num_points)
        mask_1d[sq[1:]] = 1

        self.masks_enc = mask_1d[:, None] * mask_1d[None, :] + ((1 - mask_1d[:, None]) * (1 - mask_1d[None, :]))
        self.masks_enc[0, :] = 0
        self.masks_enc[1:, 0] = 1

        self.masks_dec = self.masks_enc.copy()
        self.masks_dec[0, 0] = 1

        super().__init__([num_points, num_points], [num_points, num_points], [num_points, num_points])
        self._enc_dec_mismatch = True

    def generate_shifted_mask(self, mask_inds_vu: Union[Sequence, NDArray], mask_encoding: bool = True) -> NDArray:
        """Produce the shifted masks.

        Parameters
        ----------
        mask_inds_vu : tuple | list | NDArray
            The vertical and horizontal shifts.
        mask_encoding : bool, optional
            Is the mask encoding (False = decoding). The default is True.

        Returns
        -------
        NDArray
            The shifted mask.
        """
        if mask_encoding:
            mask = self.masks_enc
        else:
            mask = self.masks_dec

        for ii, ind in enumerate(mask_inds_vu):
            mask = np.roll(mask, ind, axis=ii)

        return self._apply_transmission(mask)

    @staticmethod
    def compute_possible_mask_sizes(fov_size: int) -> NDArray:
        """Compute MURA masks sizes.

        MURA masks require specific edge sizes: prime numbers _x_ that also
        satisfy the rule: _x_ = 4 * _l_ + 1, where _l_ is integer.

        Parameters
        ----------
        fov_size : int
            Edge size of the fov in pixels.

        Returns
        -------
        NDArray
            Array of all possible MURA mask sizes in that range.
        """

        def test_prime(x):
            div_val = x % np.arange(2, x // 2)
            return not np.any(div_val == 0)

        max_possible_val = (fov_size - 1) // 4
        test_values = np.arange(1, max_possible_val) * 4 + 1
        primes = np.array([test_prime(x) for x in test_values])
        return test_values[primes]


class ProjectorGhostImaging(operators.ProjectorOperator):
    """Projector class for the ghost imaging reconstructions."""

    mc: MaskCollection

    def __init__(self, mask_collection: Union[MaskCollection, NDArray], backend: str = "torch"):
        """
        Initialize the Ghost Imaging projector class.

        Parameters
        ----------
        mask_collection : MaskCollection
            Container of the masks.
        """
        if isinstance(mask_collection, np.ndarray):
            mask_collection = MaskCollection(mask_collection)
        self.mc = mask_collection

        self._axes_shifts = np.arange(len(self.mc.shape_shifts))
        self.col_sum = np.abs(self.mc.masks_dec).sum(axis=tuple(self._axes_shifts))
        self._axes_fov = np.arange(-len(self.mc.shape_fov), 0)
        self.row_sum = np.sqrt(np.abs(self.mc.masks_enc * self.mc.masks_dec)).sum(axis=tuple(self._axes_fov))

        self.vol_shape = self.mc.masks_enc.shape[-2:]
        self.prj_shape = np.array(self.mc.num_buckets, ndmin=1)
        super().__init__()

        if backend.lower() == "torch" and not __has_torch_cuda__:
            print("WARNING: Requested PyTorch's backend, but CUDA is not available. Reverting back to NumPy.")
            backend = "numpy"
        self.backend = backend.lower()
        self._init_backend()

    def _init_backend(self):
        if self.backend == "torch":
            self.masks_enc: pt.Tensor = pt.tensor(self.mc.masks_enc, device="cuda")
            self.masks_dec: pt.Tensor = pt.tensor(self.mc.masks_dec, device="cuda")
        elif self.backend == "numpy":
            self.masks_enc: NDArray = self.mc.masks_enc
            self.masks_dec: NDArray = self.mc.masks_dec
        else:
            raise ValueError(f"Unknown requested backend {self.backend}")

    def fp(self, image: NDArray) -> NDArray:
        """Compute forward-projection (prediction) of the bucket values.

        Parameters
        ----------
        image : NDArray
            The image for which we want to predict the bucket values.

        Returns
        -------
        NDArray
            The predicted bucket values.
        """
        masks_shape = [np.prod(self.mc.shape_shifts), np.prod(self.mc.shape_fov)]
        image_shape = [*image.shape[: -self.mc.mask_dims], np.prod(image.shape[-self.mc.mask_dims :])]

        image = image.reshape(image_shape)
        masks = self.masks_enc.reshape(masks_shape).T

        if self.backend == "numpy":
            return np.squeeze(image.dot(masks))
            # return np.sum(self.masks_enc * image, axis=(-2, -1))
        else:
            image: pt.Tensor = pt.tensor(image, device="cuda")
            bucket_vals = pt.tensordot(image, masks, ([-1], [0])).to("cpu").numpy()
            return np.squeeze(bucket_vals)

    def bp(self, bucket_vals: NDArray) -> NDArray:
        """Compute back-projection of the bucket values.

        Parameters
        ----------
        bucket_vals : NDArray
            The list of bucket values.
        subtract_mean : bool, optional
            Whether to subtract the mean of the values. The default is False.

        Returns
        -------
        NDArray
            Back-projected image.
        """
        masks_shape = [np.prod(self.mc.shape_shifts), np.prod(self.mc.shape_fov)]
        out_shape = [*bucket_vals.shape[:-1], *self.mc.shape_fov]

        masks_in = self.masks_dec.reshape(masks_shape)
        if self.backend == "numpy":
            img_out: NDArray = bucket_vals.dot(masks_in)
        else:
            buckets: pt.Tensor = pt.tensor(np.array(bucket_vals, ndmin=2), device="cuda")
            img_out: NDArray = pt.tensordot(buckets, masks_in, ([-1], [0])).to("cpu").numpy()

        return img_out.reshape(out_shape)
        # return np.sum(bucket_vals[..., None, None] * self.masks_dec, axis=-3, keepdims=True)

    def adjust_sampling_scaling(self, image: NDArray) -> NDArray:
        """Adjust reconstruction scaling and bias, due to the undersampling.

        Parameters
        ----------
        image : NDArray
            Unscaled image.

        Returns
        -------
        NDArray
            Scaled image.
        """
        sampling_ratio = self.mc.num_buckets / np.prod(self.mc.shape_fov)

        image_f: NDArray = np.fft.rfftn(image, axes=tuple(self._axes_fov), norm="ortho")
        rec_f_shape = np.array(image_f.shape)[list(self._axes_fov)]

        sampling_filter = np.ones(rec_f_shape, dtype=image.dtype) / sampling_ratio
        sampling_filter[0, 0] = 1
        image_f *= sampling_filter

        return np.fft.irfftn(image_f, s=self.mc.shape_fov, axes=tuple(self._axes_fov), norm="ortho")

    def fbp(self, bucket_vals: NDArray, use_lstsq: bool = True, adjust_scaling: bool = False) -> NDArray:
        """Compute cross-correlation reconstruction of the bucket values.

        Parameters
        ----------
        bucket_vals : NDArray
            The bucket vales to reconstruct.
        use_lstsq : bool, optional
            If True, uses least squares for reconstruction, by default True.
        adjust_scaling : bool, optional
            If True, adjusts scaling of the reconstructed image, to account for the intensity loss, by default False.

        Returns
        -------
        NDArray
            The reconstructed image.
        """
        # return self.bp(bucket_vals - np.mean(bucket_vals)) + np.mean(bucket_vals) * np.mean(self.col_sum)
        if np.any(self.mc.masks_enc != self.mc.masks_dec) and not use_lstsq:
            img = self.bp(bucket_vals)
        else:
            masks = self.mc.masks_enc.reshape([-1, np.prod(self.mc.shape_fov)])
            result = spalg.lstsq(masks, bucket_vals)
            img: NDArray = result[0]
            img = img.reshape(self.mc.shape_fov)

        if adjust_scaling:
            img = self.adjust_sampling_scaling(img)

        return img

    def absolute(self):
        """Compute the absolute value of the projection operator coefficients.

        Returns
        -------
        Op_a : ProjectorGhostImaging
            The absolute value of the projector.
        """
        Op_a = cp.deepcopy(self)
        Op_a.mc.masks_enc = np.abs(Op_a.mc.masks_enc)
        Op_a.mc.masks_dec = np.abs(Op_a.mc.masks_dec)
        Op_a._init_backend()
        return Op_a



# -*- coding: utf-8 -*-
"""
Tomographic projector backends.

@author: Nicola VIGANÒ, Computational Imaging group, CWI, The Netherlands,
and ESRF - The European Synchrotron, Grenoble, France
"""

import numpy as np
import skimage
import skimage.transform as skt

from . import filters
from .models import ProjectionGeometry, VolumeGeometry

from typing import Optional, Sequence, Union, Mapping, List
from numpy.typing import ArrayLike, NDArray

from abc import ABC, abstractmethod

try:
    import astra

    has_astra = True
    has_cuda = astra.astra.use_cuda()
    if not has_cuda:
        print(
            "WARNING: CUDA is not available. Only 2D operations on CPU are available (scikit-image will be used as default)."
        )
except ImportError:
    has_astra = False
    has_cuda = False
    print("WARNING: ASTRA is not available. Only 2D operations on CPU are available (scikit-image will be used).")

if has_cuda:
    try:
        import astra.experimental

        has_astra_direct = True
    except ImportError:
        has_astra_direct = False
        print("WARNING: the experimental ASTRA direct interface is not available. The traditional interface will be used.")


def compute_attenuation(vol: NDArray, angle_rad: float, invert: bool = False) -> NDArray:
    """
    Compute the attenuation volume for the given local attenuation, and angle.

    Parameters
    ----------
    vol : NDArray
        The local attenuation volume.
    angle_rad : float
        The angle along which to compute the attenuation.
    invert : bool, optional
        Whether to invert propagation direction. The default is False.

    Returns
    -------
    NDArray
        The attenuation volume.
    """

    def pad_vol(vol: NDArray, edges: Sequence[int]):
        paddings = np.zeros((len(vol.shape), 1), dtype=int)
        paddings[-2], paddings[-1] = edges[0], edges[1]
        return np.pad(vol, paddings, mode="constant")

    def compute_cumsum(vol: NDArray, rot_angle_deg: float):
        vol = skt.rotate(vol, -rot_angle_deg, order=1, clip=False)

        vol += np.roll(vol, 1, axis=-2)
        vol = np.cumsum(vol / 2, axis=-2)

        return skt.rotate(vol, rot_angle_deg, order=1, clip=False)

    size_lims = np.array(vol.shape[-2:])
    min_size = np.ceil(np.sqrt(np.sum(size_lims**2)))
    edges = np.ceil((min_size - size_lims) / 2).astype(int)

    if invert:
        angle_rad += np.pi

    rot_angle_deg = np.rad2deg(angle_rad)

    cum_arr = pad_vol(vol, edges)

    if cum_arr.ndim > 2:
        prev_shape = np.array(cum_arr.shape, ndmin=1)
        num_slices = np.prod(prev_shape[:-2])
        cum_arr = cum_arr.reshape([num_slices, *prev_shape[-2:]])
        for ii in range(num_slices):
            cum_arr[ii] = compute_cumsum(cum_arr[ii], rot_angle_deg)
        cum_arr = cum_arr.reshape(prev_shape)
    else:
        cum_arr = compute_cumsum(cum_arr, rot_angle_deg)
    cum_arr = cum_arr[..., edges[0] : -edges[0], edges[1] : -edges[1]]

    cum_arr = np.exp(-cum_arr)

    return cum_arr


class ProjectorBackend(ABC):
    """Base abstract projector backend class. All backends should inherit from this class."""

    vol_geom: VolumeGeometry
    vol_shape_zxy: NDArray[np.integer]

    angles_w_rad: NDArray[np.floating]

    prj_shape_vwu: NDArray[np.integer]
    prj_shape_vu: NDArray[np.integer]

    is_initialized: bool
    is_ready: bool

    def __init__(self):
        """Initialize base abstract projector backend class."""
        self.is_initialized = False
        self.is_ready = False

    def initialize_geometry(
        self,
        vol_geom: VolumeGeometry,
        angles_rot_rad: Union[ArrayLike, NDArray],
        rot_axis_shift_pix: Union[ArrayLike, NDArray, None] = None,
        prj_geom: Optional[ProjectionGeometry] = None,
        create_single_projs: bool = False,
    ):
        """Initialize the projector geometry.

        Parameters
        ----------
        vol_geom : VolumeGeometry
            The volume geometry.
        angles_rot_rad : ArrayLike | NDArray
            The projection angles.
        rot_axis_shift_pix : float | NDArray | None, optional
            Relative position of the rotation center with respect to the volume center. The default is None.
        prj_geom : ProjectionGeometry, optional
            The fully specified projection geometry.
            When active, the rotation axis shift is ignored. The default is None.
        create_single_projs : bool, optional
            Whether to create projectors for single projections. Used for corrections and SART. The default is False.
        """
        self.vol_geom = vol_geom

        self.vol_shape_zxy = self.vol_geom.shape_zxy
        self.angles_w_rad = np.array(angles_rot_rad, ndmin=1, dtype=np.float32)

        self.vol_eff_shape_zxy = np.floor(self.vol_shape_zxy * self.vol_geom.vox_size).astype(int)

        # Basic sizes, unless overridden
        self.prj_shape_vwu = np.array(
            [*self.vol_eff_shape_zxy[:-2], len(self.angles_w_rad), self.vol_eff_shape_zxy[-1]], dtype=int
        )
        self.prj_shape_vu = np.array([*self.vol_eff_shape_zxy[:-2], 1, self.vol_eff_shape_zxy[-1]], dtype=int)

        self.has_individual_projs = create_single_projs

        self.is_initialized = True

    def get_vol_shape(self) -> NDArray:
        """Return the expected and produced volume shape (in ZXY coordinates).

        Returns
        -------
        NDArray
            The volume shape.
        """
        return self.vol_shape_zxy

    def get_prj_shape(self) -> NDArray:
        """Return the expected and produced projection shape (in VWU coordinates).

        Returns
        -------
        NDArray
            The projection shape.
        """
        return self.prj_shape_vwu

    def make_ready(self) -> None:
        """Initialize the projector.

        It should make sure that all the resources have been allocated.
        """
        self.is_ready = True

    def dispose(self) -> None:
        """De-initialize the projector.

        It should make sure that all the resources have been de-allocated.
        """
        self.is_ready = False

    def __del__(self):
        """De-initialize projector on deletion."""
        if self.is_ready:
            self.dispose()

    def __repr__(self) -> str:
        """Build a string representation of the projector backend.

        Returns
        -------
        str
            The representation of the projector.
        """
        class_name = f"{self.__class__.__name__}: "
        if self.is_initialized:
            return (
                class_name
                + "{\n"
                + f"  Shape vol ZXY: {self.get_vol_shape()}\n"
                + f"  Shape prj VWU: {self.get_prj_shape()}\n"
                + f"  Angles (deg): {np.rad2deg(self.angles_w_rad)}\n"
                + "}"
            )
        else:
            return class_name + "{ Not initialized! }"

    @abstractmethod
    def fp(self, vol: NDArray, angle_ind: Optional[int] = None) -> NDArray:
        """Forward-project volume.

        Forward-projection interface. Derived backends need to implement this method.

        Parameters
        ----------
        vol : NDArray
            The volume to forward-project.
        angle_ind : int, optional
            The angle index to foward project. The default is None.
        """

    @abstractmethod
    def bp(self, prj: NDArray, angle_ind: Optional[int] = None) -> NDArray:
        """Back-project data.

        Back-projection interface. Derived backends need to implement this method.

        Parameters
        ----------
        prj : NDArray
            The sinogram to back-project or a single line.
        angle_ind : int, optional
            The angle index to foward project. The default is None.
        """


class ProjectorBackendSKimage(ProjectorBackend):
    """Projector backend based on scikit-image."""

    def __init__(self) -> None:
        """Initialize scikit-image projector backend class."""
        super().__init__()
        self.is_ready = True

    def initialize_geometry(
        self,
        vol_geom: VolumeGeometry,
        angles_rot_rad: Union[ArrayLike, NDArray],
        rot_axis_shift_pix: Union[ArrayLike, NDArray, None] = None,
        prj_geom: Optional[ProjectionGeometry] = None,
        create_single_projs: bool = False,
    ):
        """Initialize projector backend based on scikit-image.

        Parameters
        ----------
        vol_geom : VolumeGeometry
            The volume shape.
        angles_rot_rad : ArrayLike
            The projection angles.
        rot_axis_shift_pix : float | NDArray | None, optional
            Relative position of the rotation center with respect to the volume center. The default is None.
            NOT SUPPORTED: if anything else than None is passed, it will throw an error!
        prj_geom : ProjectionGeometry, optional
            The fully specified projection geometry.
            When active, the rotation axis shift is ignored. The default is None.
            NOT SUPPORTED: if anything else than None is passed, it will throw an error!
        create_single_projs : bool, optional
            Whether to create projectors for single projections. Used for corrections and SART. The default is False.

        Raises
        ------
        ValueError
            In case the volume dimensionality is larger than 2D, and if a rotation axis shift is passed.
        """
        if vol_geom.is_3D():
            raise ValueError("With the scikit-image backend, only 2D volumes are allowed!")
        if rot_axis_shift_pix is not None:
            raise ValueError("With the scikit-image backend, rotation axis shifts are not supported!")
        if prj_geom is not None:
            raise ValueError("With the scikit-image backend, `ProjectionGeometry` is not supported!")

        super().initialize_geometry(vol_geom, angles_rot_rad, create_single_projs=create_single_projs)

        self.angles_w_deg = np.rad2deg(self.angles_w_rad)

    @staticmethod
    def _set_filter_name(filt):
        if skimage.__version__ >= "0.18":
            return dict(filter_name=filt)
        else:
            return dict(filter=filt)

    @staticmethod
    def _set_bpj_size(output_size):
        return dict(circle=False, output_size=output_size)

    def fp(self, vol: NDArray, angle_ind: Optional[int] = None) -> NDArray:
        """Forward-projection of the volume to the sinogram or a single sinogram line.

        Parameters
        ----------
        vol : NDArray
            The volume to forward-project.
        angle_ind : int, optional
            The angle index to foward project. The default is None.

        Returns
        -------
        NDArray
            The forward-projected sinogram or sinogram line.
        """
        if angle_ind is None:
            prj = np.empty(self.prj_shape_vwu, dtype=vol.dtype)
            for ii_a, a in enumerate(self.angles_w_deg):
                prj[ii_a, :] = np.squeeze(skt.radon(vol, [a]))
            return prj
        else:
            return np.squeeze(skt.radon(vol, self.angles_w_deg[angle_ind : angle_ind + 1 :]))

    def bp(self, prj: NDArray, angle_ind: Optional[int] = None) -> NDArray:
        """Back-projection of a single sinogram line to the volume.

        Parameters
        ----------
        prj : NDArray
            The sinogram to back-project or a single line.
        angle_ind : int, optional
            The angle index to foward project. The default is None.

        Returns
        -------
        NDArray
            The back-projected volume.
        """
        filter_name = self._set_filter_name(None)
        bpj_size = self._set_bpj_size(self.vol_shape_zxy[-1])
        if angle_ind is None:
            vol = np.empty([self.prj_shape_vwu[-2], *self.vol_shape_zxy], dtype=prj.dtype)
            for ii_a, a in enumerate(self.angles_w_deg):
                vol[ii_a, ...] = skt.iradon(prj[ii_a, :, np.newaxis], [a], **bpj_size, **filter_name)
            vol = vol.sum(axis=0)
        else:
            vol = skt.iradon(prj[:, np.newaxis], self.angles_w_deg[angle_ind : angle_ind + 1 :], **bpj_size, **filter_name)
        return vol * 2 / np.pi


class ProjectorBackendASTRA(ProjectorBackend):
    """Projector backend based on astra-toolbox."""

    proj_id: List

    astra_vol_geom: Mapping
    proj_geom_ind: Sequence[Mapping]
    proj_geom_all: Mapping

    def __init__(self, super_sampling: int = 1):
        """Initialize the ASTRA projector backend.

        Parameters
        ----------
        super_sampling : int, optional
            Super sampling factor for the pixels and voxels, by default 1.
        """
        super().__init__()
        self.super_sampling = super_sampling

    def initialize_geometry(
        self,
        vol_geom: VolumeGeometry,
        angles_rot_rad: Union[ArrayLike, NDArray],
        rot_axis_shift_pix: Union[ArrayLike, NDArray, None] = None,
        prj_geom: Optional[ProjectionGeometry] = None,
        create_single_projs: bool = False,
    ):
        """Initialize geometry of projector backend based on astra-toolbox.

        Parameters
        ----------
        vol_geom : VolumeGeometry
            The volume shape.
        angles_rot_rad : ArrayLike
            The projection angles.
        rot_axis_shift_pix : float | NDArray | None, optional
            Relative position of the rotation center with respect to the volume center. The default is None.
        prj_geom : ProjectionGeometry, optional
            The fully specified projection geometry.
            When active, the rotation axis shift is ignored. The default is None.
        create_single_projs : bool, optional
            Whether to create projectors for single projections. Used for corrections and SART. The default is False.

        Raises
        ------
        ValueError
            In case the volume dimensionality is larger than 2D and CUDA is not available.
        """
        if vol_geom.is_3D() and not has_cuda:
            raise ValueError("CUDA is not available: only 2D volumes are allowed!")
        if not (rot_axis_shift_pix is None or isinstance(rot_axis_shift_pix, (int, float, Sequence, np.ndarray))):
            raise ValueError(
                "Rotation axis shift should either be None or one of the following: int, a float or a sequence of floats"
                + f" ({type(rot_axis_shift_pix)} given instead)."
            )

        super().initialize_geometry(vol_geom, angles_rot_rad, create_single_projs=create_single_projs)

        self.proj_id = []
        self.dispose()

        num_angles = self.angles_w_rad.size

        if self.vol_geom.is_3D():
            self.astra_vol_geom = astra.create_vol_geom(*vol_geom.shape_xyz[list([1, 0, 2])], *self.vol_geom.extent)
            if prj_geom is None:
                prj_geom = ProjectionGeometry.get_default_parallel(geom_type="3d", rot_axis_shift_pix=rot_axis_shift_pix)

            if prj_geom.det_shape_vu is None:
                prj_geom.det_shape_vu = np.array(self.prj_shape_vwu[list([-3, -1])], dtype=int)
            else:
                # Here the projections are supposed to be larger or smaller than the sample size
                self.prj_shape_vwu = np.array([*prj_geom.det_shape_vu[:-1], num_angles, prj_geom.det_shape_vu[-1]])
                self.prj_shape_vu = np.array([*prj_geom.det_shape_vu[:-1], 1, prj_geom.det_shape_vu[-1]])

            rot_geom = prj_geom.rotate(self.angles_w_rad)

            vectors = np.empty([num_angles, 12])
            # source / beam direction
            vectors[:, 0:3] = rot_geom.get_field_scaled("src_pos_xyz")
            # center of detector
            vectors[:, 3:6] = rot_geom.get_field_scaled("det_pos_xyz")
            # vector from detector pixel (0, 0) to (0, 1)
            vectors[:, 6:9] = rot_geom.get_field_scaled("det_u_xyz")
            # vector from detector pixel (0, 0) to (1, 0)
            vectors[:, 9:12] = rot_geom.get_field_scaled("det_v_xyz")

            geom_type_str = prj_geom.geom_type
        else:
            self.astra_vol_geom = astra.create_vol_geom(*vol_geom.shape_xyz[list([1, 0])], *self.vol_geom.extent)
            if prj_geom is None:
                prj_geom = ProjectionGeometry.get_default_parallel(geom_type="2d", rot_axis_shift_pix=rot_axis_shift_pix)

            if prj_geom.det_shape_vu is None:
                prj_geom.det_shape_vu = np.array(self.prj_shape_vwu[list([-1])], dtype=int)
            else:
                # Here the projections are supposed to be larger or smaller than the sample size
                self.prj_shape_vwu = np.array([num_angles, prj_geom.det_shape_vu[-1]])
                self.prj_shape_vu = np.array([1, prj_geom.det_shape_vu[-1]])

            rot_geom = prj_geom.rotate(self.angles_w_rad, patch_astra_2d=True)

            vectors = np.empty([num_angles, 6])
            # source / beam direction
            vectors[:, 0:2] = rot_geom.get_field_scaled("src_pos_xyz")
            # center of detector
            vectors[:, 2:4] = rot_geom.get_field_scaled("det_pos_xyz")
            # vector from detector pixel 0 to 1
            vectors[:, 4:6] = rot_geom.get_field_scaled("det_u_xyz")

            geom_type_str = prj_geom.geom_type[:-2]

        vectors /= vol_geom.vox_size

        if self.has_individual_projs:
            self.proj_geom_ind = [
                astra.create_proj_geom(geom_type_str + "_vec", *prj_geom.det_shape_vu, vectors[ii : ii + 1 :, :])
                for ii in range(num_angles)
            ]

        self.proj_geom_all = astra.create_proj_geom(geom_type_str + "_vec", *prj_geom.det_shape_vu, vectors)

    # def get_vol_shape(self) -> NDArray:
    #     """Return the expected and produced volume shape (in ZYX coordinates).

    #     Returns
    #     -------
    #     NDArray
    #         The volume shape.
    #     """
    #     return astra.functions.geom_size(self.astra_vol_geom)

    # def get_prj_shape(self) -> NDArray:
    #     """Return the expected and produced projection shape (in VWU coordinates).

    #     Returns
    #     -------
    #     NDArray
    #         The projection shape.
    #     """
    #     return astra.functions.geom_size(self.proj_geom_all)

    def make_ready(self) -> None:
        """Initialize the ASTRA projectors."""
        if not self.is_ready:
            if self.vol_geom.is_3D():
                projector_type = "cuda3d"
                self.algo_type = "3D_CUDA"
                self.data_mod = astra.data3d
            else:
                if has_cuda:
                    projector_type = "cuda"
                    self.algo_type = "_CUDA"
                else:
                    projector_type = "linear"
                    self.algo_type = ""
                self.data_mod = astra.data2d

            voxel_sampling = int(self.super_sampling * np.fmax(1, self.vol_geom.vox_size))
            pixel_sampling = int(self.super_sampling / np.fmin(1, self.vol_geom.vox_size))
            opts = {"VoxelSuperSampling": voxel_sampling, "DetectorSuperSampling": pixel_sampling}

            if self.has_individual_projs:
                self.proj_id = [
                    astra.create_projector(projector_type, pg, self.astra_vol_geom, opts) for pg in self.proj_geom_ind
                ]

            self.proj_id.append(astra.create_projector(projector_type, self.proj_geom_all, self.astra_vol_geom, opts))

        super().make_ready()

    def _check_data(self, x: NDArray, expected_shape: Union[Sequence[int], NDArray[np.integer]]) -> NDArray:
        if x.dtype != np.float32:
            x = x.astype(np.float32)
        if not x.flags["C_CONTIGUOUS"]:
            x = np.ascontiguousarray(x)
        try:
            return x.reshape(np.array(expected_shape, ndmin=1))
        except ValueError:
            print(f"Could not reshape input data of shape={x.shape} into expected shape={expected_shape}")
            raise

    def _check_prj_shape(self, prj: NDArray) -> None:
        inds_vu = [ind for ind in range(-len(self.prj_shape_vu), 0) if self.prj_shape_vu[ind] > 1]
        expected_prj_shape_vu = self.prj_shape_vu[inds_vu]
        given_prj_shape_vu = np.array(prj.shape)[inds_vu]
        if np.any(given_prj_shape_vu != expected_prj_shape_vu):
            raise ValueError(
                f"Expected prj shape (VU: {expected_prj_shape_vu} - likely inferred from `vol_geom`),"
                f" and passed prj shape (VU: {given_prj_shape_vu}) from the projection data, do not match.\n"
                "If you intend to use projection shapes different from the volume shape, please adjust"
                " the `prj_geom` object's `det_shape_vu` accordingly."
            )

    def dispose(self) -> None:
        """De-initialize the ASTRA projectors."""
        for p in self.proj_id:
            astra.projector.delete(p)
        self.proj_id = []

        super().dispose()

    def fp(self, vol: NDArray, angle_ind: Optional[int] = None) -> NDArray:
        """Apply forward-projection of the volume to the sinogram or a single sinogram line.

        Parameters
        ----------
        vol : NDArray
            The volume to forward-project.
        angle_ind : int | None, optional
            The angle index to forward project. The default is None.

        Returns
        -------
        NDArray
            The forward-projected sinogram or sinogram line.
        """
        self.make_ready()

        vol = self._check_data(vol, self.vol_shape_zxy)

        if angle_ind is None:
            prj = np.empty(self.prj_shape_vwu, dtype=np.float32)
            prj_geom = self.proj_geom_all
            proj_id = self.proj_id[-1]
        else:
            if not self.has_individual_projs:
                raise ValueError("Individual projectors not available!")

            prj = np.empty(self.prj_shape_vu, dtype=np.float32)
            prj_geom = self.proj_geom_ind[angle_ind]
            proj_id = self.proj_id[angle_ind]

        vid = self.data_mod.link("-vol", self.astra_vol_geom, vol)
        try:
            sid = self.data_mod.link("-sino", prj_geom, prj)
            try:
                cfg = astra.creators.astra_dict("FP" + self.algo_type)
                cfg["ProjectionDataId"] = sid
                cfg["VolumeDataId"] = vid
                cfg["ProjectorId"] = proj_id

                fp_id = astra.algorithm.create(cfg)
                try:
                    astra.algorithm.run(fp_id)
                finally:
                    astra.algorithm.delete(fp_id)
            finally:
                self.data_mod.delete([sid])
        finally:
            self.data_mod.delete([vid])

        return np.squeeze(prj)

    def bp(self, prj: NDArray, angle_ind: Optional[int] = None) -> NDArray:
        """Apply back-projection of a single sinogram line to the volume.

        Parameters
        ----------
        prj : NDArray
            The sinogram to back-project or a single line.
        angle_ind : int | None, optional
            The angle index to forward project. The default is None.

        Returns
        -------
        NDArray
            The back-projected volume.
        """
        self.make_ready()

        self._check_prj_shape(prj)

        vol = np.empty(self.vol_shape_zxy, dtype=np.float32)
        if angle_ind is None:
            prj = self._check_data(prj, self.prj_shape_vwu)
            prj_geom = self.proj_geom_all
            proj_id = self.proj_id[-1]
        else:
            if not self.has_individual_projs:
                raise ValueError("Individual projectors not available!")

            prj = self._check_data(prj, self.prj_shape_vu)
            prj_geom = self.proj_geom_ind[angle_ind]
            proj_id = self.proj_id[angle_ind]

        vid = self.data_mod.link("-vol", self.astra_vol_geom, vol)
        try:
            sid = self.data_mod.link("-sino", prj_geom, prj)

            try:
                cfg = astra.creators.astra_dict("BP" + self.algo_type)
                cfg["ProjectionDataId"] = sid
                cfg["ReconstructionDataId"] = vid
                cfg["ProjectorId"] = proj_id
                bp_id = astra.algorithm.create(cfg)
                try:
                    astra.algorithm.run(bp_id)
                finally:
                    astra.algorithm.delete(bp_id)
            finally:
                self.data_mod.delete([sid])
        finally:
            self.data_mod.delete([vid])

        return vol


class ProjectorBackendDirectASTRA(ProjectorBackendASTRA):
    """Experimental astra-toolbox functions projector."""

    astra_vol_shape: Sequence
    astra_prj_shape: Sequence
    astra_angle_prj_shape: Sequence
    angle_prj_shape: Sequence

    def initialize_geometry(
        self,
        vol_geom: VolumeGeometry,
        angles_rot_rad: Union[ArrayLike, NDArray],
        rot_axis_shift_pix: Union[ArrayLike, NDArray, None] = None,
        prj_geom: Optional[ProjectionGeometry] = None,
        create_single_projs: bool = False,
    ):
        """Initialize projector backend based on experimental astra-toolbox functions.

        Parameters
        ----------
        vol_geom : VolumeGeometry
            The volume shape.
        angles_rot_rad : ArrayLike
            The projection angles.
        rot_axis_shift_pix : float, optional
            Relative position of the rotation center with respect to the volume center. The default is 0.0.
        geom : ProjectionGeometry, optional
            The fully specified projection geometry.
            When active, the rotation axis shift is ignored. The default is None.
        create_single_projs : bool, optional
            Whether to create projectors for single projections. Used for corrections and SART. The default is False.

        Raises
        ------
        ValueError
            In case CUDA is not available.
        """
        if not has_cuda:
            raise ValueError("CUDA is not available, but it is required for the direct functions!")
        if not has_astra_direct:
            raise ValueError("ASTRA direct is not available, but it is required!")
        if not (rot_axis_shift_pix is None or isinstance(rot_axis_shift_pix, (int, float, Sequence, np.ndarray))):
            raise ValueError(
                "Rotation axis shift should either be None or one of the following: int, a float or a sequence of floats"
                + f" ({type(rot_axis_shift_pix)} given instead)."
            )

        ProjectorBackend.initialize_geometry(self, vol_geom, angles_rot_rad, create_single_projs=create_single_projs)

        self.proj_id = []
        self.dispose()

        num_angles = self.angles_w_rad.size

        vol_geom_tmp3d = self.vol_geom.get_3d()

        self.astra_vol_geom = astra.create_vol_geom(*vol_geom_tmp3d.shape_xyz[list([1, 0, 2])], *vol_geom_tmp3d.extent)
        if prj_geom is None:
            prj_geom = ProjectionGeometry.get_default_parallel(geom_type="3d", rot_axis_shift_pix=rot_axis_shift_pix)
        else:
            if prj_geom.det_shape_vu is not None:
                # Here the projections are supposed to be larger or smaller than the sample size
                # We use the original version because it makes sure that 2D is respected.
                self.prj_shape_vwu = np.array([*prj_geom.det_shape_vu[:-1], num_angles, prj_geom.det_shape_vu[-1]])
                self.prj_shape_vu = np.array([*prj_geom.det_shape_vu[:-1], 1, prj_geom.det_shape_vu[-1]])

            prj_geom = prj_geom.get_3d()

        if prj_geom.det_shape_vu is None:
            prj_shape_vu = np.delete(self.prj_shape_vwu, obj=-2)
            prj_geom.det_shape_vu = np.ones(2, dtype=int)
            prj_geom.det_shape_vu[-len(prj_shape_vu) :] = prj_shape_vu

        rot_geom = prj_geom.rotate(self.angles_w_rad)

        vectors = np.empty([num_angles, 12])
        # source / beam direction
        vectors[:, 0:3] = rot_geom.get_field_scaled("src_pos_xyz")
        # center of detector
        vectors[:, 3:6] = rot_geom.get_field_scaled("det_pos_xyz")
        # vector from detector pixel (0, 0) to (0, 1)
        vectors[:, 6:9] = rot_geom.get_field_scaled("det_u_xyz")
        # vector from detector pixel (0, 0) to (1, 0)
        vectors[:, 9:12] = rot_geom.get_field_scaled("det_v_xyz")

        geom_type_str = prj_geom.geom_type

        if self.has_individual_projs:
            self.proj_geom_ind = [
                astra.create_proj_geom(geom_type_str + "_vec", *prj_geom.det_shape_vu, vectors[ii : ii + 1 :, :])
                for ii in range(num_angles)
            ]

        self.proj_geom_all = astra.create_proj_geom(geom_type_str + "_vec", *prj_geom.det_shape_vu, vectors)

        self.astra_vol_shape = tuple(vol_geom_tmp3d.shape_zxy)
        self.astra_prj_shape = (prj_geom.det_shape_vu[-2], num_angles, prj_geom.det_shape_vu[-1])
        self.astra_angle_prj_shape = (prj_geom.det_shape_vu[-2], 1, prj_geom.det_shape_vu[-1])
        self.angle_prj_shape = (prj_geom.det_shape_vu[-2], prj_geom.det_shape_vu[-1])

    def make_ready(self):
        """Initialize the ASTRA projectors."""
        if not self.is_ready:
            voxel_sampling = int(self.super_sampling * np.fmax(1, self.vol_geom.vox_size))
            pixel_sampling = int(self.super_sampling / np.fmin(1, self.vol_geom.vox_size))
            opts = {"VoxelSuperSampling": voxel_sampling, "DetectorSuperSampling": pixel_sampling}

            if self.has_individual_projs:
                self.proj_id = [astra.create_projector("cuda3d", pg, self.astra_vol_geom, opts) for pg in self.proj_geom_ind]

            self.proj_id.append(astra.create_projector("cuda3d", self.proj_geom_all, self.astra_vol_geom, opts))

        ProjectorBackend.make_ready(self)

    def fp(self, vol: NDArray, angle_ind: Optional[int] = None):
        """Apply forward-projection of the volume to the sinogram or a single sinogram line.

        Parameters
        ----------
        vol : NDArray
            The volume to forward-project.
        angle_ind : int | None, optional
            The angle index to forward project. The default is None.

        Returns
        -------
        NDArray
            The forward-projected sinogram or sinogram line.
        """
        self.make_ready()

        if angle_ind is None:
            prj = np.zeros(self.astra_prj_shape, dtype=np.float32)
            proj_id = self.proj_id[-1]
            out_shape = self.prj_shape_vwu
        else:
            if not self.has_individual_projs:
                raise ValueError("Individual projectors not available!")
            prj = np.zeros(self.astra_angle_prj_shape, dtype=np.float32)
            proj_id = self.proj_id[angle_ind]
            out_shape = self.angle_prj_shape

        vol = self._check_data(vol, self.astra_vol_shape)
        astra.experimental.direct_FP3D(proj_id, vol, prj)
        return prj.reshape(out_shape)

    def bp(self, prj: NDArray, angle_ind: Optional[int] = None):
        """Apply back-projection of a single sinogram line to the volume.

        Parameters
        ----------
        prj : NDArray
            The sinogram to back-project or a single line.
        angle_ind : int | None, optional
            The angle index to forward project. The default is None.

        Returns
        -------
        NDArray
            The back-projected volume.
        """
        self.make_ready()

        self._check_prj_shape(prj)

        if angle_ind is None:
            prj = self._check_data(prj, self.astra_prj_shape)
            proj_id = self.proj_id[-1]
        else:
            if not self.has_individual_projs:
                raise ValueError("Individual projectors not available!")
            prj = self._check_data(prj, self.astra_angle_prj_shape)
            proj_id = self.proj_id[angle_ind]

        vol = np.zeros(self.astra_vol_shape, dtype=np.float32)
        astra.experimental.direct_BP3D(proj_id, vol, prj)
        return vol.reshape(self.vol_shape_zxy)



"""
Detector shifts finding classes.

@author: Nicola VIGANÒ, Computational Imaging group, CWI, The Netherlands,
and ESRF - The European Synchrotron, Grenoble, France
"""

from collections.abc import Mapping
from typing import Optional, Union

import matplotlib.pyplot as plt
import numpy as np
from numpy.typing import ArrayLike, NDArray
from tqdm.auto import tqdm

from . import fitting

eps = np.finfo(np.float32).eps


NDArrayFloat = NDArray[np.floating]


def _filter_shifts(shifts_vu: NDArrayFloat, max_shifts: NDArrayFloat) -> NDArrayFloat:
    invalid_shifts = np.abs(shifts_vu) > max_shifts

    shifts_vu_filt = shifts_vu.copy()
    if np.any(invalid_shifts):
        print(f"WARNING - Some shifts exceeded the maximum allowed magnitude (max: {max_shifts}). Setting them to 0.")
        shifts_vu_filt[invalid_shifts] = 0

    return shifts_vu_filt


class DetectorShiftsBase:
    """Compute the detector shifts for a given dataset."""

    data_vwu: NDArrayFloat
    angles_rad: NDArrayFloat

    def __init__(
        self,
        data_dvwu: NDArrayFloat,
        rot_angle_rad: Union[ArrayLike, NDArrayFloat],
        *,
        data_format: str = "dvwu",
        data_mask_dvwu: Optional[NDArray] = None,
        borders_dvwu: dict = {"d": None, "v": None, "w": None, "u": None},
        max_shifts: Union[float, NDArrayFloat, None] = None,
        precision_decimals: int = 2,
        verbose: bool = True,
    ):
        """Initialize the base class for detector shifts.

        Parameters
        ----------
        data_dvwu : NDArrayFloat
            The tomographic data.
        rot_angle_rad : ArrayLike | NDArrayFloat
            The rotation angles in radians.
        data_format : str, optional
            The data organization, by default "dvwu"
        data_mask_dvwu : NDArray | None, optional
            The mask of the tomographic data, by default None
        borders_dvwu : dict, optional
            The borders of the tomographic data, by default {"d": None, "v": None, "w": None, "u": None}
        max_shifts : float | NDArrayFloat | None, optional
            Maximum shifts allowed, by default None
        precision_decimals : int, optional
            The precision of the results, by default 2
        verbose : bool, optional
            Whether to be verbose, by default True

        Raises
        ------
        ValueError
            Raised when passing incoherent data and angles.
        """
        self.data_dims = len(data_dvwu.shape)

        self.data_shapes = dict(u=0, v=1, w=0, d=1)
        for ii in range(-self.data_dims, 0):
            self.data_shapes[data_format[ii]] = data_dvwu.shape[ii]

        self.num_dets = self.data_shapes["d"]

        self.angles_rad = np.array(np.squeeze(rot_angle_rad), ndmin=1, dtype=np.float32)
        if self.data_shapes["w"] != len(self.angles_rad):
            raise ValueError(
                f"Mismatch between rotation angles ({len(self.angles_rad)}),"
                f" and number of projections ({self.data_shapes['w']})."
            )

        slicing = [borders_dvwu[data_format[ii]] for ii in range(-self.data_dims, 0)]
        slicing = [slice(s, -s) if s is not None and s > 0 else slice(None) for s in slicing]
        self.slicing = tuple(slicing)

        if self.data_shapes["v"] > 1:
            self._align_coords = np.array([-3, -1])
        else:
            self._align_coords = np.array([-1])

        if max_shifts is None:
            max_shifts = np.array(data_dvwu.shape)[self._align_coords] / 2
        max_shifts = np.array(max_shifts, ndmin=1)
        self.max_shifts = max_shifts[:, None]

        self.decimals = precision_decimals
        self.verbose = verbose

        self.data_vwu = data_dvwu
        self.data_mask_vwu = data_mask_dvwu


class DetectorShiftsPRE(DetectorShiftsBase):
    """Compute the pre-alignment detector shifts for a given dataset."""

    def fit_v(
        self,
        use_derivative: bool = True,
        use_rfft: bool = True,
        normalize_fourier: bool = True,
    ) -> NDArrayFloat:
        """Compute the pre-alignment vertical shifts of a 3D dataset.

        The pre-alignment shifts are computed by cross-correlation of one
        projection against the others.
        The projections are integrated in the horizontal direction.

        In the vertical direction, it is suggested to use some high pass filter.
        The default option is to use the derivates of the intensity profiles.

        Parameters
        ----------
        use_derivative : bool, optional
            Whether to use the derivate of the vertical profile, by default True
        use_rfft : bool, optional
            Whether to use the `rfft` transform for the cross-correlation, by default True
        normalize_fourier : bool, optional
            Whether to normalize the cross-correlation in Fourier space, by default True

        Returns
        -------
        NDArrayFloat
            The vertical shifts.

        Raises
        ------
        ValueError
            If the dataset is 2D.
        """
        if self.data_shapes["v"] <= 1:
            raise ValueError("Vertical alignment not supported for 2D reconstructions.")

        if use_rfft:
            local_fft = np.fft.rfft
            local_ifft = np.fft.irfft
        else:
            local_fft = np.fft.fft
            local_ifft = np.fft.ifft

        data_vwu = self.data_vwu[self.slicing]
        if self.data_mask_vwu is None:
            data_vw = np.mean(data_vwu, axis=-1)
        else:
            data_mask_vwu = self.data_mask_vwu[self.slicing].astype(data_vwu.dtype)
            data_vwu = data_vwu * data_mask_vwu
            mask_sum_vw = data_mask_vwu.sum(axis=-1)
            data_vw = data_vwu.sum(axis=-1) / (mask_sum_vw + (mask_sum_vw == 0))
        if self.num_dets > 1:
            data_vw = np.mean(data_vw, axis=-3)

        mins = data_vw.min(axis=-2, keepdims=True)
        maxs = data_vw.max(axis=-2, keepdims=True)
        data_vw_d = (data_vw - mins) / (maxs - mins)

        pad_size = ((data_vw.shape[-2] // 2,), (0,))
        if use_derivative:
            data_vw_d = np.diff(data_vw_d, axis=-2)
            data_vw_p = np.pad(data_vw_d, pad_width=pad_size, mode="constant")
        else:
            data_vw_p = np.pad(data_vw_d, pad_width=pad_size, mode="linear_ramp")

        data_vw_f = local_fft(data_vw_p, axis=-2)

        ref_angle = len(self.angles_rad) // 2
        ccs_f = data_vw_f[:, [ref_angle]] * data_vw_f.conj()

        if normalize_fourier:
            ccs_f /= np.fmax(np.abs(ccs_f).max(axis=-2, keepdims=True), eps)

        cross_corr = local_ifft(ccs_f, axis=-2).real

        cc_coords = np.fft.fftfreq(cross_corr.shape[-2], 1 / cross_corr.shape[-2])
        f_vals, fc_ax = fitting.extract_peak_regions_1d(cross_corr, axis=-2, cc_coords=cc_coords)
        shifts_v = fitting.refine_max_position_1d(f_vals, decimals=self.decimals) + fc_ax[1, :]

        shifts_v = _filter_shifts(shifts_v, self.max_shifts[0, :])

        shifts_v -= np.mean(shifts_v)
        shifts_v = np.around(shifts_v, decimals=self.decimals)

        if self.verbose:
            fig, axs = plt.subplots(2, 2, figsize=[10, 5], sharex=True)
            axs[0, 0].imshow(data_vw)
            axs[0, 0].set_title("Data VW")
            axs[0, 0].set_xlabel("Coord. W (angular)")
            axs[0, 0].set_ylabel("Coord. V (vertical)")
            axs[0, 1].plot(shifts_v)
            axs[0, 1].plot(np.zeros_like(shifts_v))
            axs[0, 1].grid()
            axs[0, 1].set_title(f"Shifts V (wrt angle n.{ref_angle})")
            axs[0, 1].set_xlabel("Coord. W (angular)")
            axs[0, 1].set_ylabel("Coord. V (vertical)")
            axs[1, 0].imshow(data_vw_p)
            axs[1, 0].set_title("Data used for cross-correlation")
            axs[1, 1].imshow(np.fft.fftshift(cross_corr, axes=(-2,)))
            axs[1, 1].set_title("Cross-correlation")
            fig.tight_layout()
            plt.show(block=False)

        return shifts_v

    def fit_u(
        self,
        fit_l1: bool = False,
        background: Union[float, NDArray, None] = None,
        method: str = "com",
    ) -> tuple[NDArrayFloat, float]:
        """Compute the pre-alignment shifts for the horizontal dimension.

        The pre-alignment shifts, and center-of-rotation (CoR) are computed by
        fitting a sinusoid to the centers of mass of each angle in the sinogram.
        The bias of the sinusoid corresponds to the CoR, while the deviations
        from the fitted curve correspond to the shifts.

        Parameters
        ----------
        fit_l1 : bool, optional
            Computes the l1-min fit of the sinusoid, by default False.
        background : float | NDArray | None, optional
            Removes the given background, by default None.
        method : str, optional
            The method used for the identification of the fiducial marker position.
            Options are "com" (center-of-mass) | "max" (maximum value), by default "com".

        Returns
        -------
        Tuple[NDArrayFloat, float]
            The shifts and the CoR.
        """
        is_3d = self.data_shapes["v"] > 1

        data_vwu = self.data_vwu[self.slicing]

        if self.num_dets > 1:
            data_vwu = np.mean(data_vwu, axis=0)

        if is_3d:
            data_vwu = np.mean(data_vwu, axis=-3)

        if background is not None:
            data_vwu = data_vwu - background
        data_vwu = np.fmax(data_vwu, 0.0)

        fx_half_size = (data_vwu.shape[-1] - 1) / 2

        if method.lower() == "com":
            fx = np.linspace(-fx_half_size, fx_half_size, data_vwu.shape[-1])
            ref_points = -np.sum(data_vwu * fx, axis=-1) / np.sum(data_vwu, axis=-1)
        elif method.lower() == "max":
            ref_points = fx_half_size - np.argmax(data_vwu, axis=-1)
        else:
            raise ValueError(f"Unknown selected method {method}. Please choose one among: 'com' | 'max'")

        a, p, b = fitting.fit_sinusoid(self.angles_rad, ref_points, fit_l1=fit_l1)

        cor = np.around(b, decimals=self.decimals)

        if self.verbose:
            angles_deg = np.rad2deg(self.angles_rad)
            sort_angles_deg = np.sort(angles_deg)
            sort_angles_rad = np.sort(self.angles_rad)
            fig, axs = plt.subplots(1, 2, figsize=(10, 5))
            axs[0].scatter(angles_deg, ref_points, label="Centers of mass")
            axs[0].plot(sort_angles_deg, fitting.sinusoid(sort_angles_rad, a, p, b), c="C1", label="Fitted sinusoid")
            axs[0].plot(sort_angles_deg, np.ones_like(sort_angles_deg) * b, c="C2", label="Bias (CoR)")
            axs[0].legend()
            axs[0].grid()
            axs[0].set_xlabel("Coord. W (angular)")
            axs[0].set_ylabel("Coord. U (horizontal)")
            axs[0].xaxis.label.set_fontsize(16)
            axs[0].yaxis.label.set_fontsize(16)
            axs[1].imshow(data_vwu)
            axs[1].scatter(-ref_points + fx_half_size, np.arange(len(ref_points)), c="C1")
            axs[1].set_xlabel("Coord. U (horizontal)")
            axs[1].set_ylabel("Coord. W (angular)")
            axs[1].xaxis.label.set_fontsize(16)
            axs[1].yaxis.label.set_fontsize(16)
            fig.tight_layout()
            plt.show(block=False)
            print(f"amplitude = {a} (pix)")
            print(f"phase = {p} (rad)")
            print(f"bias = {b} (pix)")
            print(f" -> cor = {cor} (pix)")

        shifts_u: NDArrayFloat = ref_points - fitting.sinusoid(self.angles_rad, a, p, b)
        shifts_u = _filter_shifts(shifts_u, self.max_shifts[-1, :])
        shifts_u = np.around(shifts_u, decimals=self.decimals)

        return shifts_u, float(cor)


class DetectorShiftsXC(DetectorShiftsBase):
    """Compute the center-of-rotation for a given dataset, by cross correlation."""

    def fit_vu_accum_drifts(self, ref_data_dvwu: Optional[NDArrayFloat] = None) -> NDArray:
        """Fit static image drifts.

        Parameters
        ----------
        ref_data_dvwu : Optional[NDArrayFloat], optional
            Reference image, by default None. If None, the first image in the
            data stack will be used.

        Returns
        -------
        NDArray
            The shifts of the image stack.

        Raises
        ------
        ValueError
            When the number of reference images is either too many or not enough.
        """
        if ref_data_dvwu is None:
            ref_data_dvwu = self.data_vwu[..., [0], :]

        img_inds = np.arange(self.data_shapes["w"])
        if ref_data_dvwu.shape[-2] == 1:
            ref_inds = np.zeros_like(img_inds)
        else:
            ref_inds = np.arange(ref_data_dvwu.shape[-2])
        if img_inds.size != ref_inds.size:
            raise ValueError(
                f"Reference images should either be 1 or as many as the data images"
                f" ({img_inds.size}), but {ref_inds.size} were passed instead"
            )

        is_3d = self.data_shapes["v"] > 1
        num_dims = 1 + is_3d
        rel_shifts_vu = np.zeros((num_dims, img_inds.size))
        for ii, (ind_ref, ind_img) in enumerate(zip(tqdm(ref_inds, desc="Computing drifts"), img_inds)):
            rel_shifts_vu[..., [ii]] = self.find_shifts_vu(ref_data_dvwu[..., [ind_ref], :], self.data_vwu[..., [ind_img], :])

        return rel_shifts_vu

    def fit_vu(self, fit_l1: bool = False) -> NDArray:
        """Compute the pre-alignment vertical and horizontal shifts, using cross-correlation.

        Parameters
        ----------
        fit_l1 : bool, optional
            Computes the l1-min fit of the sinusoid, by default False.

        Returns
        -------
        NDArray
            Pre-alignment shifts in VU coordinates.
        """
        is_3d = self.data_shapes["v"] > 1
        num_dims = 1 + is_3d

        angles_order = np.argsort(self.angles_rad)
        sorted_angles_rad = self.angles_rad[angles_order]

        rel_shifts_vu = np.zeros((num_dims, len(self.angles_rad)))
        desc = "Computing adjacent images shifts"
        for ii, (a_prev, a_curr) in enumerate(zip(tqdm(angles_order[:-1], desc=desc), angles_order[1:])):
            rel_shifts_vu[..., [ii + 1]] = self.find_shifts_vu(
                self.data_vwu[..., [a_prev], :], self.data_vwu[..., [a_curr], :]
            )

        shifts_vu = np.cumsum(rel_shifts_vu, axis=-1)
        if is_3d:
            shifts_vu[0, :] -= np.mean(shifts_vu[0, :])
            shifts_vu[0, :] = np.around(shifts_vu[0, :], decimals=self.decimals)

        a, p, b = fitting.fit_sinusoid(sorted_angles_rad, shifts_vu[-1, ...], fit_l1=fit_l1)

        if self.verbose:
            coords = ["V", "U"] if is_3d else ["U"]
            fig, axs = plt.subplots(1, num_dims, figsize=[10, 5], sharex=True, squeeze=False)
            angles_deg = np.rad2deg(sorted_angles_rad)
            for ii, (s, coord) in enumerate(zip(shifts_vu, coords)):
                axs[0, ii].plot(angles_deg, s, label="Shifts")
                if coord == "V":
                    axs[0, ii].plot(angles_deg, np.zeros_like(s), label="Zero")
                else:
                    axs[0, ii].plot(angles_deg, np.ones_like(s) * b, label="Bias")
                    axs[0, ii].plot(angles_deg, fitting.sinusoid(sorted_angles_rad, a, p, b), label="Expected motion")
                axs[0, ii].legend()
                axs[0, ii].grid()
                axs[0, ii].set_title(f"Shifts {coord}")
            fig.tight_layout()
            plt.show(block=False)

        shifts_u: NDArrayFloat = shifts_vu[-1, ...] - fitting.sinusoid(sorted_angles_rad, a, p, b)
        shifts_u = _filter_shifts(shifts_u, self.max_shifts[-1, :])
        shifts_u = np.around(shifts_u, decimals=self.decimals)

        shifts_vu[-1] = shifts_u

        return shifts_vu[:, list(angles_order)]

    def fit_u_180(self) -> float:
        """Find the center-of-rotation, using the 0 and 180 degrees projections.

        Returns
        -------
        float
            The center-of-rotation.
        """
        angle_0 = self.angles_rad.min()
        angle_0_ind = np.argmin(np.abs(self.angles_rad - angle_0))
        angle_180_ind = np.argmin(np.abs(self.angles_rad - (angle_0 + np.pi)))

        angle_180 = self.angles_rad[angle_180_ind]
        if not np.isclose(angle_0 + np.pi, angle_180):
            print(
                f"WARNING - No opposite angles found ({np.rad2deg(angle_0)} and {np.rad2deg(angle_180)})."
                " Center-of-rotation will be inaccurate."
            )
        img_0 = self.data_vwu[..., [angle_0_ind], :]
        img_180 = np.flip(self.data_vwu[..., [angle_180_ind], :], axis=-1)
        # upsample_factor = 1 / (10 ** (-self.decimals))
        # return skr.phase_cross_correlation(img_0, img_180, upsample_factor=upsample_factor, return_error=False)
        shifts_vu = self.find_shifts_vu(img_0, img_180)
        return -shifts_vu[-1] / 2

    def fit_u_360(self) -> float:
        """Find the center of rotation over a 360 degrees scan, by taking the average of the 0-180 over all pairs of angles.

        Returns
        -------
        float
            The center-of-rotation.
        """
        # We should be checking whether the scan is really 360 or not.

        angles_boundary = self.angles_rad[0] + np.pi
        num_angles = np.sum(self.angles_rad < angles_boundary)

        a1s = self.angles_rad[:num_angles]
        a2s = a1s + np.pi

        iis_1 = np.arange(num_angles)
        iis_2 = np.argmin(np.abs(self.angles_rad[None, :] - a2s[:, None]), axis=-1)

        shifts_vu = np.empty([2, len(iis_1)])
        for ii, (ii1, ii2) in enumerate(tqdm(zip(iis_1, iis_2), total=num_angles)):
            img_1 = self.data_vwu[..., [ii1], :]
            img_2 = np.flip(self.data_vwu[..., [ii2], :], axis=-1)
            shifts_vu[..., [ii]] = self.find_shifts_vu(img_1, img_2)

        cors = -shifts_vu[-1, :] / 2

        # upsample_factor = int(1 / (10 ** (-self.decimals)))

        # cors = np.empty_like(a1s)
        # for ii_1, a1 in enumerate(tqdm(a1s)):
        #     ii_2 = np.argmin(np.abs(self.angles_rad - (a1 + np.pi)))
        #     # We should be handling non-redundant scans.

        #     img_0 = self.data_vwu[..., ii_1, :]
        #     img_180 = np.flip(self.data_vwu[..., ii_2, :], axis=-1)
        #     s = skr.phase_cross_correlation(img_0, img_180, upsample_factor=upsample_factor, return_error=False)

        #     cors[ii_1] = -s[-1] / 2

        cor = np.around(np.mean(cors), decimals=self.decimals)

        if self.verbose:
            fig, axs = plt.subplots(1, 1, figsize=[10, 5])
            axs.plot(cors)
            axs.plot(np.ones_like(cors) * cor)
            axs.grid()
            axs.set_title("Centers of rotation")
            fig.tight_layout()
            plt.show(block=False)

        return float(cor)

    def find_shifts_vu(
        self,
        data_dvwu: NDArrayFloat,
        proj_dvwu: NDArrayFloat,
        use_derivative: bool = False,
        xc_opts: Mapping = dict(normalize_fourier=False),
    ) -> NDArrayFloat:
        """Find shifts between two images or sets of lines.

        Parameters
        ----------
        data_dvwu : NDArrayFloat
            The reference data.
        proj_dvwu : NDArrayFloat
            The other data.
        use_derivative : bool, optional
            Whether to use derivatives over the horizontal (U) coordinate, by default False.

        Returns
        -------
        NDArrayFloat
            The shifts in vertical (optional) and horizontal coordinates ([V]U).
        """
        if self.num_dets == 1:
            data_dvwu = data_dvwu[None, ...]
            proj_dvwu = proj_dvwu[None, ...]

        shifts_vu_all = [np.array([])] * self.num_dets

        for ii_d in range(self.num_dets):
            data_vwu = data_dvwu[ii_d]
            proj_vwu = proj_dvwu[ii_d]

            if use_derivative:
                data_vwu = np.diff(data_vwu, axis=-1)
                proj_vwu = np.diff(proj_vwu, axis=-1)

            # Allow to choose different shift finding functions
            shifts_vu = fitting.fit_shifts_vu_xc(data_vwu, proj_vwu, decimals=self.decimals, **xc_opts)
            # shifts_vu = fitting.fit_shifts_u_sad(data_vwu, proj_vwu, decimals=self.decimals)

            shifts_vu_all[ii_d] = _filter_shifts(shifts_vu, self.max_shifts)

        return np.mean(shifts_vu_all, axis=0)



#!/usr/bin/env python3
"""
Phase contrast support functions.

@author: Nicola VIGANÒ, CEA-IRIG, Grenoble, France
"""

try:
    from . import xraylib_helper  # noqa: F401, F402

    xraylib = xraylib_helper.xraylib

except ImportError:
    print("WARNING: Physics support is only available when xraylib is installed!")
    raise

from collections.abc import Sequence
from typing import Union
import matplotlib.pyplot as plt
import numpy as np
import scipy.ndimage as spimg
from numpy.typing import DTypeLike, NDArray


def get_delta_beta(cmp_name: str, energy_keV: float, density: Union[float, None] = None) -> float:
    """Compute the delta-over-beta parameter for a specific compound.

    Parameters
    ----------
    cmp_name : str
        Molar composition of a compound
    energy_keV : float
        Energy at which the d/b value should be computed
    density : Optional[float], optional
        Density of the compound, by default None

    Returns
    -------
    float
        The computed delta-over-beta.
    """
    cmp = xraylib_helper.get_compound(cmp_name, density=density)
    ref_ind = xraylib.Refractive_Index(cmp_name, energy_keV, cmp["density"])
    return (1 - ref_ind.real) / ref_ind.imag


def get_delta_beta_curves(
    compounds: Sequence[str], energy_keV_range: tuple[float, float, int] = (1.0, 800.0, 500), plot: bool = True
) -> Sequence[NDArray]:
    """Compute the delta-over-beta curves for the listed compounds, in the requested energy range.

    Parameters
    ----------
    compounds : Sequence[str]
        Sequence of compounds.
    energy_keV_range : tuple[float, float, int], optional
        The energy range in keV, by default (1.0, 800.0, 500)
    plot : bool, optional
        Whether to plot the result, by default True

    Returns
    -------
    Sequence[NDArray]
        The computed delta-over-beta valueas.
    """
    energies = np.linspace(*energy_keV_range)
    delta_betas_cmps = []
    for cmp in compounds:
        delta_betas_cmp = np.array([get_delta_beta(cmp, e) for e in energies])
        delta_betas_cmps.append(delta_betas_cmp)

    if plot:
        fig, axs = plt.subplots(1, 1)
        for ii, cmp in enumerate(compounds):
            axs.plot(energies, delta_betas_cmps[ii], label=cmp)
        axs.grid()
        axs.legend(fontsize=13)
        axs.set_xlabel("Energy (keV)", fontsize=14)
        axs.set_ylabel("$\\delta/\\beta$", fontsize=14)
        axs.tick_params(labelsize=13)
        fig.tight_layout()
        plt.show(block=False)

    return delta_betas_cmps


def _tie_freq_response(k2: NDArray, dist_um: float, wlength_um: float, delta_beta: float) -> NDArray:
    return 1 + delta_beta * dist_um * wlength_um * np.pi * k2


def _ctf_freq_response(k2: NDArray, dist_um: float, wlength_um: float, delta_beta: float) -> NDArray:
    steps = dist_um * wlength_um * np.pi * k2
    return np.cos(steps) + delta_beta * np.sin(steps)


def plot_filter_responses(
    filter_length: int, pix_size_um: float, dist_um: float, wlength_um: float, delta_beta: float, domain: str = "fourier"
) -> tuple:
    """Plot frequency response of the wave propagation.

    Parameters
    ----------
    filter_length : int
        Length of the filter in pixels
    pix_size_um : float
        Pixel-wise of the detector in microns
    dist_um : float
        Distance of the detector from sample in microns
    wlength_um : float
        Wavelength of the wave in microns
    delta_beta : float
        Radio between the refraction index decrement and the absorption coefficient.
    domain : str
        Whether to plot Fourier or direct-space responses. By default, "Fourier".

    Returns
    -------
    tuple
        Figure and axes of the plot
    """
    k = np.fft.rfftfreq(filter_length, d=pix_size_um)
    k2 = k**2
    tie_resp = _tie_freq_response(k2, dist_um=dist_um, wlength_um=wlength_um, delta_beta=delta_beta)
    ctf_resp = _ctf_freq_response(k2, dist_um=dist_um, wlength_um=wlength_um, delta_beta=delta_beta)
    if domain.lower() == "fourier":
        step = k
    elif domain.lower() == "direct":
        tie_resp = np.fft.fftshift(np.fft.irfft(tie_resp))
        ctf_resp = np.fft.fftshift(np.fft.irfft(ctf_resp))
        p = np.fft.fftfreq(filter_length, d=1.0 / (pix_size_um * filter_length))
        p = np.fft.fftshift(p)
        step = p
    else:
        raise ValueError(f"Unknown domain {domain}, please choose one of 'Fourier' | 'direct'.")

    fig, axs = plt.subplots(1, 1, figsize=[8, 4])
    axs.set_title(f"Domain: {domain}")
    axs.plot(step, tie_resp, label="TIE")
    axs.plot(step, ctf_resp, label="CTF")
    axs.grid()
    axs.legend()
    fig.tight_layout()

    return fig, axs


def get_propagation_filter(
    img_shape: Union[Sequence[int], NDArray],
    pix_size_um: float,
    dist_um: float,
    wlength_um: float,
    delta_beta: float,
    filter_type: str = "ctf",
    use_rfft: bool = False,
    plot_result: bool = False,
    dtype: DTypeLike = np.float32,
) -> tuple[NDArray, NDArray]:
    """Compute the phase contrast propagation filter for the given parameters.

    Parameters
    ----------
    img_shape : Union[Sequence[int], NDArray]
        Shape of the target image
    pix_size_um : float
        Pixel size of the detector (in microns)
    dist_um : float
        Propagation distance (in microns)
    wlength_um : float
        Wavelength of the radiation (in microns)
    delta_beta : float
        Delta-over-beta value for the given material
    filter_type : str, optional
        Type of the filter, by default "ctf".
        Options: "ctf" (Contrast Transfer Function) | "tie" (Transport of Intensity Equation)
    use_rfft : bool, optional
        Whether to use the rfft, by default False
    plot_result : bool, optional
        Whether to plot the result, by default False
    dtype : DTypeLike, optional
        Data type to use, by default np.float32

    Returns
    -------
    tuple[NDArray, NDArray]
        The Fourier-space and real-space filters, respectively

    Raises
    ------
    ValueError
        When choosing an incorrect filter type
    """
    if use_rfft:
        coords = [np.fft.fftfreq(s, d=pix_size_um) for s in img_shape[:-1]] + [np.fft.rfftfreq(img_shape[-1], d=pix_size_um)]
    else:
        coords = [np.fft.fftfreq(s, d=pix_size_um) for s in img_shape]
    coords = np.meshgrid(*coords, indexing="ij")
    k2 = np.sum(np.stack(coords, axis=0) ** 2, axis=0)

    if filter_type.lower() == "ctf":
        filt_fourier = _ctf_freq_response(k2=k2, dist_um=dist_um, wlength_um=wlength_um, delta_beta=delta_beta)
    elif filter_type.lower() == "tie":
        filt_fourier = _tie_freq_response(k2=k2, dist_um=dist_um, wlength_um=wlength_um, delta_beta=delta_beta)
    else:
        raise ValueError(f"Invalid filter type: {filter_type}. Possible choices: 'ctf' | 'tie'")

    fft_axes = tuple(np.arange(-len(img_shape), 0))
    if use_rfft:
        filt_direct = np.fft.irfftn(filt_fourier, s=tuple(img_shape), axes=fft_axes)
    else:
        filt_direct = np.fft.ifftn(filt_fourier, s=tuple(img_shape), axes=fft_axes).real

    if plot_result:
        fig, axs = plt.subplots(
            2, 1, sharex=True, sharey=True, figsize=tuple(np.flip(img_shape) * 12 / np.max(img_shape) * np.array([1, 2]))
        )
        axs[0].imshow(filt_fourier)
        axs[1].imshow(np.fft.fftshift(filt_direct))
        fig.tight_layout()

    return filt_fourier.astype(dtype), np.fft.fftshift(filt_direct.astype(dtype))


def apply_propagation_filter(
    data_wvu: NDArray, pix_size_um: float, dist_um: float, wlength_um: float, delta_beta: float, filter_type: str = "tie"
) -> NDArray:
    """Apply a requested propagation filter to an image or a stack of images.

    Parameters
    ----------
    data_wvu : NDArray
        The Image or stack of images
    pix_size_um : float
        Pixel size of the detector (in microns)
    dist_um : float
        Propagation distance (in microns)
    wlength_um : float
        Wavelength of the radiation (in microns)
    delta_beta : float
        Delta-over-beta value for the given material
    filter_type : str, optional
        Type of the filter, by default "tie".
        Options: "ctf" (Contrast Transfer Function) | "tie" (Transport of Intensity Equation)

    Returns
    -------
    NDArray
        The filtered image
    """
    pad_width = [(0, 0) for _ in range(data_wvu.ndim)]
    pad_size = np.array(data_wvu.shape[-2:])
    pad_width[-2] = pad_size[-2] - pad_size[-2] // 2, pad_size[-2] // 2
    pad_width[-1] = pad_size[-1] - pad_size[-1] // 2, pad_size[-1] // 2
    data_p = np.pad(data_wvu, pad_width=pad_width, mode="edge")
    data_f = np.fft.rfftn(data_p, axes=(-2, -1))

    filt_f, _ = get_propagation_filter(
        data_p.shape[-2:],
        pix_size_um=pix_size_um,
        dist_um=dist_um,
        wlength_um=wlength_um,
        delta_beta=delta_beta,
        filter_type=filter_type,
        use_rfft=True,
    )

    data_f = spimg.fourier_shift(
        data_f,
        shift=(*([0] * (data_wvu.ndim - 2)), pad_size[-2] // 2 - pad_size[-2], pad_size[-1] // 2 - pad_size[-1]),
        n=data_p.shape[-1],
    )
    data_f /= filt_f

    return np.fft.irfftn(data_f, axes=(-2, -1))[..., : data_wvu.shape[-2], : data_wvu.shape[-1]]



# -*- coding: utf-8 -*-

"""Package for data pre-processing and post-processing."""

__author__ = """Nicola VIGANÒ"""
__email__ = "N.R.Vigano@cwi.nl"

from . import pre  # noqa: F401, F402
from . import post  # noqa: F401, F402
from . import misc  # noqa: F401, F402
from . import noise  # noqa: F401, F402

import numpy as np

circular_mask = misc.circular_mask

# Functions to compute the variance of signals
compute_variance_poisson = noise.compute_variance_poisson
compute_variance_transmission = noise.compute_variance_transmission

# Function to compute the weights associated to the given variance
compute_variance_weight = noise.compute_variance_weight



#!/usr/bin/env python3
"""
Test `corrct.projectors` package.

@author: Nicola VIGANÒ, Computational Imaging group, CWI, The Netherlands,
and ESRF - The European Synchrotron, Grenoble, France
"""

from typing import Callable, Optional

import matplotlib.pyplot as plt
import numpy as np
import pytest
import scipy.ndimage as spndi
import skimage.data as skd
import skimage.transform as skt
from numpy.typing import NDArray

import corrct as cct
from corrct import _projector_backends as backends

eps = np.finfo(np.float32).eps


def _radon_rot_sk_w(x: NDArray, angles_rad: NDArray, shift: Optional[float] = None) -> NDArray:
    """
    Compute the Radon transform using skimage's warp function.

    Parameters
    ----------
    x : NDArray
        Input image.
    angles_rad : NDArray
        Angles in radians.
    shift : float, optional
        Shift value. Default is None.

    Returns
    -------
    NDArray
        Radon transform of the input image.
    """
    prj = np.empty((len(angles_rad), x.shape[-1]), dtype=np.float32)
    c = (np.array(x.shape) - 1) / 2

    I = np.eye(3, dtype=np.float32)
    C = np.array([[0, 0, c[1]], [0, 0, c[0]], [0, 0, 0]], dtype=np.float32)
    T1 = I - C
    T2 = I + C
    if shift is not None:
        S = np.array([[0, 0, shift], [0, 0, 0], [0, 0, 0]], dtype=np.float32)
    else:
        S = np.zeros_like(I)

    for ii, angle in enumerate(angles_rad):
        cos_a, sin_a = np.cos(angle), np.sin(angle)
        R = np.array([[cos_a, sin_a, 0], [-sin_a, cos_a, 0], [0, 0, 1]], dtype=np.float32)

        # Transformation matrix, that takes the center into account
        T2_R_T1 = T2.dot(R.dot(T1))
        rotated_x = skt.warp(x, T2_R_T1, clip=False)
        if shift is not None:
            rotated_x = skt.warp(rotated_x, (I + S), clip=False)

        prj[ii, ...] = rotated_x.sum(axis=0)
    return prj


def _radon_rot_sp(x: NDArray, angles_rad: NDArray) -> NDArray:
    """
    Compute the Radon transform using scipy's rotate function.

    Parameters
    ----------
    x : NDArray
        Input image.
    angles_rad : NDArray
        Angles in radians.

    Returns
    -------
    NDArray
        Radon transform of the input image.
    """
    prj = np.empty((len(angles_rad), x.shape[-1]), dtype=np.float32)
    for ii, a in enumerate(np.rad2deg(angles_rad)):
        prj[ii, ...] = spndi.rotate(x, -a, order=1, reshape=False).sum(axis=0)
    return prj


@pytest.fixture(scope="function")
def base_test_data() -> tuple[NDArray, NDArray, float]:
    """
    Fixture to provide base test data.

    Returns
    -------
    tuple[NDArray, NDArray, float]
        Phantom image, angles in radians, and shift value in pixels.
    """
    ph = skd.shepp_logan_phantom()
    angles_deg = np.arange(0, 180)
    angles_rad = angles_deg / 180 * np.pi
    shift = 20
    return ph, angles_rad, shift


def _test_centered_sinogram(ph: NDArray, angles_rad: NDArray, ref_function: Callable) -> None:
    """
    Test the centered sinogram.

    Parameters
    ----------
    ph : NDArray
        Phantom image.
    angles_rad : NDArray
        Angles in radians.
    ref_function : callable
        Reference function to compute the Radon transform.
    """
    debug = False

    ph_added = ph + 0.1 * cct.processing.circular_mask(ph.shape)

    prj_ref = ref_function(ph_added, angles_rad)

    with cct.projectors.ProjectorUncorrected(ph.shape, angles_rad, rot_axis_shift_pix=None) as prj:
        prj_cmp = prj(ph_added)

        assert prj.projector_backend.__class__.__name__ in [
            "ProjectorBackendASTRA",
            "ProjectorBackendDirectASTRA",
        ], "The projector backend is not Astra's or Direct Astra's. This function will not work as intended"

    rel_diff = (prj_ref - prj_cmp) / prj_ref.max()

    if debug:
        print(np.max(np.abs(prj_ref - prj_cmp)))
        print(np.max(np.abs(rel_diff)))
        fig, axs = plt.subplots(3, 1)
        axs[0].imshow(prj_ref)
        axs[1].imshow(prj_cmp)
        axs[2].imshow(prj_ref - prj_cmp)
        fig.tight_layout()
        plt.show()

    assert np.all(np.isclose(rel_diff, 0, atol=0.015)), "Reference radon transform and astra-toolbox do not match"


@pytest.mark.skipif(not cct.projectors.astra_available, reason="astra-toolbox not available")
def test_centered_sinogram_sk(base_test_data: tuple[NDArray, NDArray, float]) -> None:
    """
    Test the centered sinogram using skimage's warp function.

    Parameters
    ----------
    base_test_data : tuple[NDArray, NDArray, float]
        Base test data containing:
        - Phantom image (NDArray)
        - Angles in radians (NDArray)
        - Shift value in pixels (float)
    """
    ph, angles_rad, _ = base_test_data
    _test_centered_sinogram(ph, angles_rad, _radon_rot_sk_w)


@pytest.mark.skipif(not cct.projectors.astra_available, reason="astra-toolbox not available")
def test_centered_sinogram_sp(base_test_data: tuple[NDArray, NDArray, float]) -> None:
    """
    Test the centered sinogram using scipy's rotate function.

    Parameters
    ----------
    base_test_data : tuple[NDArray, NDArray, float]
        Base test data containing:
        - Phantom image (NDArray)
        - Angles in radians (NDArray)
        - Shift value in pixels (float)
    """
    ph, angles_rad, _ = base_test_data
    _test_centered_sinogram(ph, angles_rad, _radon_rot_sp)


@pytest.mark.skipif(not cct.projectors.astra_available, reason="astra-toolbox not available")
def test_shifted_sinogram(base_test_data: tuple[NDArray, NDArray, float]) -> None:
    """
    Test the shifted sinogram.

    Parameters
    ----------
    base_test_data : tuple[NDArray, NDArray, float]
        Base test data containing:
        - Phantom image (NDArray)
        - Angles in radians (NDArray)
        - Shift value in pixels (float)
    """
    ph, angles_rad, shift = base_test_data
    debug = False

    ph_added = ph + 0.1 * cct.processing.circular_mask(ph.shape)

    prj_ref = _radon_rot_sk_w(ph_added, angles_rad, shift)

    with cct.projectors.ProjectorUncorrected(ph.shape, angles_rad, rot_axis_shift_pix=shift) as A:
        prj_cmp = A(ph_added)

    rel_diff = (prj_ref - prj_cmp) / prj_ref.max()

    if debug:
        print(np.max(np.abs(prj_ref - prj_cmp)))
        print(np.max(np.abs(rel_diff)))
        fig, axs = plt.subplots(3, 1)
        axs[0].imshow(prj_ref)
        axs[1].imshow(prj_cmp)
        axs[2].imshow(prj_ref - prj_cmp)
        fig.tight_layout()
        plt.show()

    assert np.all(np.isclose(rel_diff, 0, atol=0.015)), "Reference radon transform and astra-toolbox do not match"


@pytest.mark.skipif(not cct.projectors.astra_available, reason="astra-toolbox not available")
def test_astra_backends(base_test_data: tuple[NDArray, NDArray, float]) -> None:
    """
    Test the astra backends.

    Parameters
    ----------
    base_test_data : tuple[NDArray, NDArray, float]
        Base test data containing:
        - Phantom image (NDArray)
        - Angles in radians (NDArray)
        - Shift value in pixels (float)
    """
    ph, angles_rad, shift = base_test_data
    debug = False

    projector_legacy = backends.ProjectorBackendASTRA()
    projector_direct = backends.ProjectorBackendDirectASTRA()

    solver = cct.solvers.SIRT()
    num_iterations = 25

    prj_ref = _radon_rot_sk_w(ph, angles_rad, shift)

    with cct.projectors.ProjectorUncorrected(ph.shape, angles_rad, rot_axis_shift_pix=shift, backend=projector_legacy) as A:
        rec_vol_leg, _ = solver(A, prj_ref, iterations=num_iterations)

    with cct.projectors.ProjectorUncorrected(ph.shape, angles_rad, rot_axis_shift_pix=shift, backend=projector_direct) as A:
        rec_vol_dir, _ = solver(A, prj_ref, iterations=num_iterations)

    diff_rec = rec_vol_leg - rec_vol_dir
    rel_diff = diff_rec / rec_vol_leg.max()

    if debug:
        print(np.max(np.abs(diff_rec)))
        print(np.max(np.abs(rel_diff)))
        fig, ax = plt.subplots(1, 3, figsize=[9, 3])
        ax[0].imshow(rec_vol_leg)
        ax[1].imshow(rec_vol_dir)
        ax[2].imshow(diff_rec)
        fig.tight_layout()
        plt.show()

    assert np.allclose(rel_diff, 0.0, atol=0.0002), "Legacy and direct astra-toolbox projectors do not match"
